from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule

def lineno():
    """Returns the current line number in our program."""
    return str(' - ManagedPolicyOnUserRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class ManagedPolicyOnUserRule(BaseRule):
  
  
  def __init__(self, cfn_model=None, debug=None):
    """
    Initialize
    :param cfn_model: 
    """
    BaseRule.__init__(self, cfn_model, debug=debug)

  def rule_text(self):
    """
    Get rule text
    :return: 
    """
    if self.debug:
      print('rule_text'+lineno())
    return 'IAM managed policy should not apply directly to users.  Should be on group'


  def rule_type(self):
    """
    Get rule type
    :return: 
    """
    self.type= 'VIOLATION::FAILING_VIOLATION'
    return 'VIOLATION::FAILING_VIOLATION'

  def rule_id(self):
    """
    Get rule id
    :return: 
    """
    if self.debug:
      print('rule_id'+lineno())
    self.id ='F12'
    return 'F12'


  def audit_impl(self):
    """
    Audit
    :return: violations 
    """
    if self.debug:
      print('ManagedPolicyOnUserRule - audit_impl'+lineno())

    violating_policies = []

    resources= self.cfn_model.resources_by_type('AWS::IAM::ManagedPolicy')

    if len(resources)>0:
      for resource in resources:
          if self.debug:
            print('resource: '+str(resource)+lineno())

          if hasattr(resource,'users'):
            if resource.users:
              if self.debug:
                print('users: '+str(resource.users))
              if len(resource.users)>0:

                violating_policies.append(str(resource.logical_resource_id))

    else:
      if self.debug:
        print('no violating_policies' + lineno())

    if self.debug:
      input('Press enter to continue: '+lineno())

    return violating_policies


import os
import sys
import unittest
from collections import OrderedDict
from terraform_validator.ValidateUtility import ValidateUtility as class_to_test

def pretty(value, htchar='\t', lfchar='\n', indent=0):
    """
    Prints pretty json
    :param value:
    :param htchar:
    :param lfchar:
    :param indent:
    :return: pretty json
    """


    nlch = lfchar + htchar * (indent + 1)
    if type(value) == type(dict()) or type(value) == type(OrderedDict()):
        items = [
            nlch + repr(key) + ': ' + pretty(value[key], htchar, lfchar, indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + lfchar + htchar * indent)

    elif type(value) == type(list()):
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]

        if items:
            items = sorted(items)
        [str(item) for item in items]
        return '[%s]' % (','.join(items) + lfchar + htchar * indent)

    elif type(value) is tuple:
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + lfchar + htchar * indent)

    else:
        return repr(str(value))

class TestSqs(unittest.TestCase):
    """
    Test sqs
    """
    def test_sqs_policy(self):

          expected_result =  [
            {
                'failure_count': '0',
                'filename': '/tf/sqs_queue_policy/sqs_policy_with_not_action.tf',
                'file_results': [
                    {
                        'id': 'W18',
                        'type': 'VIOLATION::WARNING',
                        'message': 'SQS Queue policy should not allow Allow+NotAction',
                        'logical_resource_ids': "['QueuePolicyWithNotAction', 'QueuePolicyWithNotAction2']"
                    }
                ]
            }
        ]

          if sys.version_info[0] < 3:
              new_file_results = []

              for info in expected_result[0]['file_results']:
                  print('info: ' + str(info))
                  print('type: ' + str(type(info)))
                  order_of_keys = ["id", "type", "message", "logical_resource_ids"]

                  new_results = OrderedDict()
                  for key in order_of_keys:
                      new_results[key] = info[key]

                  new_file_results.append(new_results)
                  print('new file results: ' + str(new_file_results))

                  expected_result[0]['file_results'] = new_file_results

              order_of_keys = ["failure_count", "filename", "file_results"]
              list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
              expected_result = [OrderedDict(list_of_tuples)]

          expected_result = pretty(expected_result)

          template_name = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))+'/terraform_validator/test_templates/tf/sqs_queue_policy/sqs_policy_with_not_action.tf'
          debug = False

          config_dict = {}
          config_dict['template_file'] = template_name
          config_dict['debug'] = debug
          config_dict['profile'] = None
          config_dict['rules_directory'] = None
          config_dict['input_path'] = None
          config_dict['profile'] = None
          config_dict['allow_suppression'] = False
          config_dict['print_suppression'] = False
          config_dict['parameter_values_path'] = None
          config_dict['isolate_custom_rule_exceptions'] = None
          validator = class_to_test(config_dict)

          real_result =  validator.validate()
          self.maxDiff = None
          self.assertEqual(expected_result, real_result)




from __future__ import absolute_import, division, print_function
import re
import inspect
from terraform_validator.Profile import Profile
from terraform_model.parser.ParserError import ParserError


def lineno():
    """Returns the current line number in our program."""
    return str(' - ProfileLoader - line number: '+str(inspect.currentframe().f_back.f_lineno))



# Load rule profile
class ProfileLoader:
    """
    Profile loader
    """

    def __init__(self,rules_registry,debug=False):
        """
        Initialize the ProfileLoader
        :param rules_registry:
        :param debug:
        """
        self.debug = debug
        self.rules_registry = rules_registry

        if self.debug:
            print('ProfileLoader - init'+lineno())

    def load(self, profile_definition):
        """
        Load rules from a profile definition
        :param profile_definition:
        :return:
        """
        if self.debug:
            print('load'+lineno())
            print('vars: '+str(vars(profile_definition))+lineno())

        # coerce falsy profile_definition into empty string for
        # empty profile check

        if not profile_definition:
            raise ParserError("Empty profile")

        new_profile = Profile()

        if self.debug:
            print('vars: '+str(vars(new_profile))+lineno())

        for definition in profile_definition.rules:
            if self.debug:
                print('definition: '+str(definition))
            rule_id = self.rule_line_match(definition.id)

            if rule_id:
                self.check_valid_rule_id(rule_id)
                new_profile.add_rule(rule_id)

        return new_profile


    def rule_line_match(self, rule_id):
        """
        Parses a line, returns first matching line or false if no match
        :param rule_id:
        :return:
        """
        if self.debug:
            print('rule_line_match'+lineno())
            print('rule_id: '+str(rule_id)+lineno())

        rule_id = rule_id.rstrip('\r\n')

        matchObj = re.match(r'^([a-zA-Z]*?[0-9]+)\s*(.*)', rule_id, re.M | re.I)

        if matchObj:
            if self.debug:
                print("matchObj.group() : "+str(matchObj.group()))
                print("matchObj.group(1) : "+str(matchObj.group(1)))
                print("matchObj.group(2) : "+str(matchObj.group(2)))
            return matchObj.group(1)
        else:
            if self.debug:
                print("No match!!")
            return False


    def rules_ids(self):
        """
        Returns ids of rules in registry
        :return:
        """
        if self.debug:
            print('rules_ids'+lineno())

        ids = []

        for rules in self.rules_registry:
            ids.append(rules.id)




    def check_valid_rule_id(self, rule_id):
        """
        Returns true if rule_id is valid (present in rules registry), else raise an error
        :param rule_id:
        :return:
        """
        if self.debug:
            print('check_valid_rule_id'+lineno())
            print('rule_id: '+str(rule_id)+lineno())

        if self.rules_registry.by_id(rule_id) == None:

            return False

        return True


import os
import sys
import unittest
from collections import OrderedDict
from terraform_validator.ValidateUtility import ValidateUtility as class_to_test

def pretty(value, htchar='\t', lfchar='\n', indent=0):
    """
    Prints pretty json
    :param value:
    :param htchar:
    :param lfchar:
    :param indent:
    :return: pretty json
    """


    nlch = lfchar + htchar * (indent + 1)
    if type(value) == type(dict()) or type(value) == type(OrderedDict()):
        items = [
            nlch + repr(key) + ': ' + pretty(value[key], htchar, lfchar, indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + lfchar + htchar * indent)

    elif type(value) == type(list()):
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]

        if items:
            items = sorted(items)
        [str(item) for item in items]
        return '[%s]' % (','.join(items) + lfchar + htchar * indent)

    elif type(value) is tuple:
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + lfchar + htchar * indent)

    else:
        return repr(str(value))

class TestSnsPolicy(unittest.TestCase):
    """
    Test sns policy
    """
    def test_sns_policy(self):

        expected_result =  [
            {
                'failure_count': '4',
                'filename': '/tf/sns_topic_policy/sns_topic_with_wildcard_principal.tf',
                'file_results': [
                    {
                        'id': 'F18',
                        'type': 'VIOLATION::FAILING_VIOLATION',
                        'message': 'SNS topic policy should not allow * principal',
                        'logical_resource_ids': "['mysnspolicy0', 'mysnspolicy1', 'mysnspolicy2', 'mysnspolicy3']"
                    }
                ]
            }
        ]

        if sys.version_info[0] < 3:

            expected_result = [
                    {
                        'failure_count': '4',
                        'filename': '/tf/sns_topic_policy/sns_topic_with_wildcard_principal.tf',
                        'file_results': [
                            {
                                'id': 'F18',
                                'type': 'VIOLATION::FAILING_VIOLATION',
                                'message': 'SNS topic policy should not allow * principal',
                                'logical_resource_ids': "['mysnspolicy0', 'mysnspolicy1', 'mysnspolicy2', 'mysnspolicy3']"
                            }
                        ]
                    }
            ]
            new_file_results = []


            for info in expected_result[0]['file_results']:
              print('info: '+str(info))
              print('type: '+str(type(info)))
              order_of_keys = ["id", "type", "message","logical_resource_ids"]

              new_results = OrderedDict()
              for key in order_of_keys:
                  new_results[key]=info[key]

              new_file_results.append(new_results)
              print('new file results: '+str(new_file_results))

              expected_result[0]['file_results']=new_file_results


            order_of_keys = ["failure_count", "filename", "file_results"]
            list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
            expected_result = [OrderedDict(list_of_tuples)]


        expected_result = pretty(expected_result)

        template_name = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))+'/terraform_validator/test_templates/tf/sns_topic_policy/sns_topic_with_wildcard_principal.tf'
        debug = False

        config_dict = {}
        config_dict['template_file'] = template_name
        config_dict['debug'] = debug
        config_dict['profile'] = None
        config_dict['rules_directory'] = None
        config_dict['input_path'] = None
        config_dict['profile'] = None
        config_dict['allow_suppression'] = False
        config_dict['print_suppression'] = False
        config_dict['parameter_values_path'] = None
        config_dict['isolate_custom_rule_exceptions'] = None
        validator = class_to_test(config_dict)

        real_result =  validator.validate()
        self.maxDiff = None

        print('expected results: ' + str(expected_result))
        print('real results: ' + str(real_result))

        self.assertEqual(expected_result, real_result)



from __future__ import absolute_import, division, print_function
import inspect
import sys


def lineno():
    """Returns the current line number in our program."""
    return str(' - ModelElement - line number: '+str(inspect.currentframe().f_back.f_lineno))


class ModelElement():
    """
    Model element
    """
    def __init__(self, cfn_model, debug=False):
        """
        Initialize
        :param cfn_model:
        :param debug:
        """
        # attr_accessor :logical_resource_id, :resource_type, :metadata
        self.logical_resource_id = None
        self.resource_type = None
        self.metadata = None
        self.debug= debug

        self.cfn_model = cfn_model
        if self.debug:
            print('ModelElement - init'+lineno())

    def another_element(self, another_model_element):
        """
        ???
        :param another_model_element:
        :return:
        """
        if self.debug:
            print('another element'+lineno())
        # FIXME
        sys.exit(1)
        #found_unequal_instance_var = false
        #instance_variables_without_at_sign.each do |instance_variable|
        #  if instance_variable != :logical_resource_id && instance_variable != :cfn_model
        #    if self.send(instance_variable) != another_model_element.send(instance_variable)
        #      found_unequal_instance_var = true
        #    end
        #  end
        #end
        #!found_unequal_instance_var

    def method_missing(self, method_name, *args):
        """
        ???
        :param method_name:
        :param args:
        :return:
        """
        if self.debug:
            print('method_missing'+lineno())
        # FIXME
        sys.exit(1)
        #if method_name =~ / ^ (\w+)=$ /
        #instance_variable_set
        #"@#{$1}", args[0]

        #else
        #References.resolve_value( @ cfn_model, instance_variable_get("@#{method_name}"))
        #end

    def instance_variables_without_at_sign(self):
        """
        Instance variables without an at sign
        :return:
        """
        if self.debug:
            print('instance_variables_without_at_sign'+lineno())
        # FIXME
        sys(exit)
        #self.instance_variables.map { |instance_variable| strip(instance_variable) }

    def strip(self, sym):
        """
        ???
        :param sym:
        :return:
        """
        if self.debug:
            print('strip'+lineno())

        # FIXME
        sys.exit(1)
        #sym.to_s.gsub( / @ /, '').to_sym

    def emit_instance_vars(self):
        """
        ???
        :return:
        """
        if self.debug:
            print('emit_instance_vars'+lineno())

        instance_vars_str = ''

        sys.exit(1)
        #for variable in self.instance_variables_without_at_sign()
        #self.instance_variables.each do |instance_variable|
        #  instance_vars_str += "  #{instance_variable}=#{instance_variable_get(instance_variable)}\n"

        #return instance_vars_str


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule
from terraform_validator.IpAddr import IpAddr

def lineno():
    """Returns the current line number in our program."""
    return str(' - SecurityGroupEgressOpenToWorldRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class SecurityGroupEgressOpenToWorldRule(BaseRule):
  

  def __init__(self, cfn_model=None, debug=None):
    """
    Initialize
    :param cfn_model: 
    """
    BaseRule.__init__(self, cfn_model, debug=debug)

  def rule_text(self):
    """
    Get rule text
    :return: 
    """
    if self.debug:
        print('rule_text'+lineno())
    return 'Security Groups found with cidr open to world on egress'


  def rule_type(self):
    """
    Get rule type
    :return: 
    """
    self.type= 'VIOLATION::WARNING'
    return 'VIOLATION::WARNING'


  def rule_id(self):
    """
    Get rule id
    :return: 
    """
    if self.debug:
        print('rule_id'+lineno())
    self.id ='W5'
    return 'W5'


  ##
  # This will behave slightly different than the legacy jq based rule which was targeted against inline ingress only
  def audit_impl(self):
    """
    Audit
    :return: violations 
    """
    if self.debug:
        print('SecurityGroupEgressOpenToWorldRule - audit_impl'+lineno())
    violating_egresses = []

    for groups in self.cfn_model.security_groups():
        if self.debug:
            print('group: '+str(groups)+lineno())
            print('vars: '+str(vars(groups))+lineno())

        for egress in groups.egresses:
            if self.debug:
                print('egress: '+str(egress)+lineno())

            if IpAddr.ip4_open(egress, debug=self.debug) or IpAddr.ip6_open(egress, debug=self.debug):
              if self.debug:
                print('ip4/6 address is open'+lineno())
              violating_egresses.append(str(groups.logical_resource_id))

    routes= self.cfn_model.standalone_egress()

    if self.debug:
        print('routes: '+str(routes)+lineno())
    for standalone_egress in routes:
      if self.debug:
        print('standalone_egress: '+str(standalone_egress)+lineno())
        print('vars: '+str(vars(standalone_egress))+lineno())

      if IpAddr.ip4_open(standalone_egress,debug=self.debug) or IpAddr.ip6_open(standalone_egress,debug=self.debug):
        if self.debug:
            print('ip4/6 address is open' + lineno())
        violating_egresses.append(standalone_egress.logical_resource_id)

    if self.debug:
      input('Press enter to continue: '+lineno())

    return violating_egresses



from __future__ import absolute_import, division, print_function
import inspect
import sys

def lineno():
    """Returns the current line number in our program."""
    return str(' -  Ec2NetworkInterfaceParser- line number: '+str(inspect.currentframe().f_back.f_lineno))


class Ec2NetworkInterfaceParser:
    """
    Ec2 network interface parser
    """
    @staticmethod
    def parse(cfn_model, resource, debug=False):
        print('Ec2NetworkInterfaceParser - parse'+lineno())
        # FIXME
        sys.exit(1)
    #network_interface = resource

    #if network_interface.groupSet.is_a? Array
    #  network_interface.security_groups = network_interface.groupSet.map do |security_group_reference|
    #    cfn_model.find_security_group_by_group_id(security_group_reference)
    #  end
    #else
    #  network_interface.security_groups = []
    #end

    #network_interface


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule


def lineno():
    """Returns the current line number in our program."""
    return str(' -  CloudFrontDistributionAccessLoggingRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



# Rule class to ensure a CF distribution has logging
class CloudFrontDistributionAccessLoggingRule(BaseRule):

    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize CloudFrontDistributionAccessLoggingRule
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Returns rule text
        :return:
        """
        if self.debug:
          print('rule_text'+lineno())
        return 'CloudFront Distribution should enable access logging'


    def rule_type(self):
        """
        Returns rule type
        :return:
        """
        self.type= 'VIOLATION::WARNING'
        return 'VIOLATION::WARNING'

    def rule_id(self):
        """
        Returns the rule id
        :return:
        """
        if self.debug:
            print('rule_id'+lineno())
        self.id = 'W10'
        return 'W10'

    def audit_impl(self):
        """
        Audit
        :return: violations
        """
        if self.debug:
            print('CloudFrontDistributionAccessLoggingRule - audit_impl'+lineno())

        violating_distributions = []
        resources = self.cfn_model.resources_by_type('AWS::CloudFront::Distribution')

        if len(resources)>0:
            for resource in resources:
                if self.debug:
                    print('resource: '+str(resource)+lineno())

                if hasattr(resource, 'distributionConfig'):
                    if self.debug:
                        print('has distributionConfig ' + lineno())

                        print(resource.distributionConfig)
                    if 'Logging' not in resource.distributionConfig:
                        if self.debug:
                            print('does not have logging')

                        violating_distributions.append(str(resource.logical_resource_id))

        else:
            if self.debug:
                print('no violating_distributions'+lineno())

        if self.debug:
            input('Press enter to continue: ' + lineno())

        return violating_distributions


from __future__ import absolute_import, division, print_function
import pkg_resources
from terraform_validator.ValidateUtility import ValidateUtility #noqa

__version__ = pkg_resources.get_distribution('terraform_validator').version

__all__ = [
    'schema_registry',
    'rules_set_registry'
]
__title__ = 'terraform_validator'
__version__ = '0.6.40'
__author__ = 'Will Rubel'
__author_email__ = 'willrubel@gmail.com'




import os
import sys
import unittest
from mock import patch
from collections import OrderedDict
from contextlib import contextmanager
if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO
    import io
from terraform_validator.CustomRuleLoader import CustomRuleLoader
from terraform_validator.additional_custom_rules.Ec2CustomTagsRule import Ec2CustomTagsRule
from terraform_validator.RuleRegistry import RuleRegistry
from terraform_validator.result_views.RulesView import RulesView as class_to_test

def pretty(value, htchar='\t', lfchar='\n', indent=0):
    nlch = lfchar + htchar * (indent + 1)
    if type(value) == type(dict()) or type(value) == type(OrderedDict()):

        items = [
            nlch + repr(key) + ': ' + pretty(value[key], htchar, lfchar, indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) == type(list()):
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '[%s]' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) is tuple:
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + lfchar + htchar * indent)
    else:
        return repr(value)


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class TestRulesView(unittest.TestCase):
    """
    Test rules view
    """

    def test_emit(self):

      expected_result ="##################################\n\
########## WARNINGS ##############\n\
##################################\n\
##################################\n\
########## FAILINGS #############\n\
##################################\n\
{\"id\": \"F86\", \"type\": \"VIOLATION::FAILING_VIOLATION\", \"message\": \"EC2 instance does not have the required tags of Name, ResourceOwner, DeployedBy, Project\"}"

      rule = Ec2CustomTagsRule()

      registry = RuleRegistry(debug=False)

      registry.definition('F86', 'VIOLATION::FAILING_VIOLATION', 'EC2 instance does not have the required tags of Name, ResourceOwner, DeployedBy, Project')

      registry.add_rule(violation_def=rule)

      with captured_output() as (out, err):
          class_to_test.emit(rule_registry=registry, profile=None, debug=False)
      # This can go inside or outside the `with` block
      output = out.getvalue().strip()


      self.maxDiff = None

      print('expected results: ' + str(expected_result))
      print('real results: ' + str(output))

      self.assertEqual(expected_result, output)





from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule

def lineno():
    """Returns the current line number in our program."""
    return str(' - S3BucketPolicyNotActionRule- caller: '+str(inspect.stack()[1][3])+'  - line number: '+str(inspect.currentframe().f_back.f_lineno))



class S3BucketPolicyNotActionRule(BaseRule):
  
  def __init__(self, cfn_model=None, debug=None):
    """
    Initialize
    :param cfn_model: 
    """
    BaseRule.__init__(self, cfn_model, debug=debug)
      
  def rule_text(self):
    """
    Get rule text
    :return: 
    """
    if self.debug:
      print('rule_text'+lineno())
    return 'S3 Bucket policy should not allow Allow+NotAction'


  def rule_type(self):
    """
    Get rule type
    :return: 
    """
    self.type= 'VIOLATION::WARNING'
    return 'VIOLATION::WARNING'


  def rule_id(self):
    """
    Get rule id
    :return: 
    """
    if self.debug:
      print('rule_id'+lineno())
    self.id ='W20'
    return 'W20'


  def audit_impl(self):
    """
    Audit
    :return: violations 
    """
    if self.debug:
      print('S3BucketPolicyNotActionRule - audit_impl'+lineno())

    violating_policies = []
    resources = self.cfn_model.resources_by_type('AWS::S3::BucketPolicy')

    if len(resources) > 0:
      for resource in resources:
        if self.debug:
          print('resource: ' + str(resource) + lineno())

        if hasattr(resource, 'policy_document'):

          if resource.policy_document:
            if self.debug:
              print('has policy document ' + lineno())

            if self.debug:
              print(resource.policy_document.statements)
            if resource.policy_document.allows_not_action():
              if self.debug:
                print('has allows not action'+lineno())

              violating_policies.append(str(resource.logical_resource_id))

    else:
      if self.debug:
        print('no violating_policies' + lineno())

    if self.debug:
      input('Press enter to continue: '+lineno())

    return violating_policies


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule


def lineno():
    """Returns the current line number in our program."""
    return str(' - UserMissingGroupRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))


class UserMissingGroupRule(BaseRule):

    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Get rule text
        :return:
        """
        if self.debug:
          print('rule_text'+lineno())
        return 'User is not assigned to a group'


    def rule_type(self):
        """
        Get rule type
        :return:
        """
        self.type= 'VIOLATION::FAILING_VIOLATION'
        return 'VIOLATION::FAILING_VIOLATION'


    def rule_id(self):
        """
        Get rule id
        :return:
        """
        if self.debug:
          print('rule_id'+lineno())
        self.id ='F2000'
        return 'F2000'


    def audit_impl(self):
        """
        Audit
        :return: violations
        """
        if self.debug:
            print('UserMissingGroupRule - audit_impl'+lineno())
        logical_resource_ids = []

        for user in self.cfn_model.iam_users():
            if self.debug:
                print('user: '+str(user))
                print('vars: '+str(vars(user)))

            if hasattr(user, 'group_names'):
                if self.debug:
                    print('hass group_names attribute'+lineno())

                if len(user.group_names)<1:

                    if self.debug:
                        print('no group names')

                    logical_resource_ids.append(str(user.logical_resource_id))

        if self.debug:
            input('Press enter to continue: ' + lineno())

        return logical_resource_ids




from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule

def lineno():
    """Returns the current line number in our program."""
    return str(' - WafWebAclDefaultActionRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))


class WafWebAclDefaultActionRule(BaseRule):

  def __init__(self, cfn_model=None, debug=None):
    """
    Initialize
    :param cfn_model: 
    """
    BaseRule.__init__(self, cfn_model, debug=debug)
      
  def rule_text(self):
    """
    Get rule text
    :return: 
    """
    if self.debug:
      print('rule_text'+lineno())
    return 'WebAcl DefaultAction should not be ALLOW'


  def rule_type(self):
    """
    Get rule type
    :return: 
    """
    self.type= 'VIOLATION::FAILING_VIOLATION'
    return 'VIOLATION::FAILING_VIOLATION'


  def rule_id(self):
    """
    Get rule id
    :return: 
    """
    if self.debug:
      print('rule_id'+lineno())
    self.id ='F665'
    return 'F665'


  def audit_impl(self):
    """
    Audit
    :return: violations 
    """
    if self.debug:
      print('WafWebAclDefaultActionRule - audit_impl'+lineno())
    
    violating_web_acls = []
    resources = self.cfn_model.resources_by_type('AWS::WAF::WebACL')

    if len(resources)>0:
      for resource in resources:
          if self.debug:
            print('resource: '+str(resource)+lineno())

          if hasattr(resource,'defaultAction'):
            if resource.defaultAction:
              if self.debug:
                print('defaultAction: '+str(resource.defaultAction))
              if resource.defaultAction['Type'] == 'ALLOW':

                violating_web_acls.append(str(resource.logical_resource_id))

    else:
      if self.debug:
        print('no violating_web_acls' + lineno())

    if self.debug:
      input('Press enter to continue: '+lineno())

    return violating_web_acls


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule

def lineno():
    """Returns the current line number in our program."""
    return str(' - SnsTopicPolicyNotActionRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class SnsTopicPolicyNotActionRule(BaseRule):

    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Get rule text
        :return:
        """
        if self.debug:
            print('rule_text'+lineno())
        return 'SNS Topic policy should not allow Allow+NotAction'


    def rule_type(self):
        """
        Get rule type
        :return:
        """
        self.type= 'VIOLATION::WARNING'
        return 'VIOLATION::WARNING'


    def rule_id(self):
        """
        Get rule id
        :return:
        """
        if self.debug:
            print('rule_id'+lineno())
        self.id ='W19'
        return 'W19'


    def audit_impl(self):
        """
        Audit
        :return: violations
        """
        if self.debug:
            print('SnsTopicPolicyNotActionRule - audit_impl'+lineno())

        violating_policies = []

        resources = self.cfn_model.resources_by_type('AWS::SNS::TopicPolicy')

        if len(resources)>0:
              for resource in resources:
                  if self.debug:
                      print('resource: '+str(resource)+lineno())
                      print('vars: '+str(vars(resource))+lineno())

                  if hasattr(resource,'policy_document'):

                      if resource.policy_document:
                            if self.debug:
                                print('policy_document: '+str(resource.policy_document))
                            if resource.policy_document.allows_not_action():

                                violating_policies.append(str(resource.logical_resource_id))

        else:
            if self.debug:
                print('no violating_policies' + lineno())

        if self.debug:
            input('Press enter to continue: ' + lineno())

        return violating_policies


from __future__ import absolute_import, division, print_function
from terraform_model.model.ModelElement import ModelElement


class IAMGroup(ModelElement):
    """
    IAM group model
    """
    def __init__(self, cfn_model):
        """
        Initialize
        :param cfn_model: 
        """
        ModelElement.__init__(self, cfn_model)

        self.managedPolicyArns= []
        self.policies= []
        self.policy_objects= []

        self.resource_type = 'AWS::IAM::Group'






from __future__ import absolute_import, division, print_function
import sys
import inspect
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule

def lineno():
    """Returns the current line number in our program."""
    return str(' - IamManagedPolicyNotResourceRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class IamManagedPolicyNotResourceRule(BaseRule):

    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize IamManagedPolicyNotResourceRule
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Returns rule text
        :return:
        """
        return 'IAM managed policy should not allow Allow+NotResource'


    def rule_type(self):
        """
        Returns rule type
        :return:
        """
        self.type= 'VIOLATION::WARNING'
        return 'VIOLATION::WARNING'


    def rule_id(self):
        """
        Returns rule id
        :return:
        """
        if self.debug:
            print('rule_id'+lineno())
        self.id ='W23'
        return 'W23'


    def audit_impl(self):
        """
        Audit
        :return: violations
        """
        if self.debug:
            print('IamManagedPolicyNotResourceRule - audit_impl'+lineno())

        violating_policies = []

        resources = self.cfn_model.resources_by_type('AWS::IAM::ManagedPolicy')

        if len(resources)>0:
            for resource in resources:
                if self.debug:
                    print('resource: '+str(resource))

                if hasattr(resource,'policy_document'):
                    if self.debug:
                        print('has policy document '+lineno())

                        print('policy document: '+str(resource.policy_document)+lineno())
                    if resource.policy_document:
                        if resource.policy_document.allows_not_resource():
                            if self.debug:
                                print('has allows not resource')

                            violating_policies.append(str(resource.logical_resource_id))

                else:
                    if self.debug:
                        print('does not have policy document'+lineno())

        else:
            if self.debug:
                print('no violating_policies' + lineno())

        if self.debug:
            input('Press enter to continue: ' + lineno())

        return violating_policies


from __future__ import absolute_import, division, print_function
from terraform_model.model.ModelElement import ModelElement


class IAMManagedPolicy(ModelElement):

    """
    IAM managed policy model
    """
    def __init__(self, cfn_model):
        """
        Initialize
        :param cfn_model: 
        """
        ModelElement.__init__(self,cfn_model)

        self.groups = []
        self.roles = []
        self.users = []
        self.policy_document=None
        self.resource_type= 'AWS::IAM::ManagedPolicy'


    def policy_document(self, document):
        """
        Set the policy document
        :param document: 
        :return: 
        """
        self.policy_document=document



from __future__ import absolute_import, division, print_function
import sys
import inspect
import json
import traceback
from dill.source import getname
from terraform_model.validator import CloudformationValidator
from terraform_model.parser import TransformRegistry
from terraform_model.validator import ReferenceValidator
from terraform_model.model import CfnModel
from terraform_model.model import IAMManagedPolicy
from terraform_model.model import IAMRole
from terraform_model.model import IAMUser
from terraform_model.model import IAMGroup
from terraform_model.model import IAMPolicy
from terraform_model.model import EC2Instance
from terraform_model.model import EC2SecurityGroup
from terraform_model.model import EC2SecurityGroupEgress
from terraform_model.model import EC2SecurityGroupIngress
from terraform_model.model import ElasticLoadBalancingLoadBalancer
from terraform_model.model import ElasticLoadBalancingV2LoadBalancer
from terraform_model.model import S3BucketPolicy
from terraform_model.model import SNSTopicPolicy
from terraform_model.model import SQSQueuePolicy
from terraform_model.model import ModelElement
from terraform_model.model.Parameter import Parameter
from terraform_model.model.Data import Data
from terraform_model.model.Locals import Locals
from terraform_model.model.Output import Output
from terraform_model.model.Provider import Provider
from terraform_model.parser.ParserRegistry import ParserRegistry
from terraform_model.parser.SecurityGroupParser import SecurityGroupParser
from terraform_model.parser.Ec2NetworkInterfaceParser import Ec2NetworkInterfaceParser
from terraform_model.parser.Ec2InstanceParser import Ec2InstanceParser
from terraform_model.parser.LoadBalancerParser import LoadBalancerParser
from terraform_model.parser.LoadBalancerV2Parser import LoadBalancerV2Parser
from terraform_model.parser.IamGroupParser import IamGroupParser
from terraform_model.parser.IamUserParser import IamUserParser
from terraform_model.parser.IamRoleParser import IamRoleParser
from terraform_model.parser.WithPolicyDocumentParser import WithPolicyDocumentParser
from terraform_model.model.Policy import Policy
from terraform_model.model.PolicyDocument import PolicyDocument
from terraform_model.parser.PolicyDocumentParser import PolicyDocumentParser
from terraform_model.parser.ParserError import ParserError


def lineno():
    """Returns the current line number in our program."""
    return str(' - CfnParser - line number: '+str(inspect.currentframe().f_back.f_lineno))


class CfnParser:
    """
    Cloudformation parser
    """
    def __init__(self, debug=False):
        """
        Initialize
        :param debug: 
        """
        self.debug = debug

        if self.debug:
            print('CfnParser - init'+lineno())


    # Given raw json/tf CloudFormation template, returns a CfnModel object
    # or raise ParserErrors if something is amiss with the format
    def parse(self,cloudformation_yml, parameter_values_json=None):

        if self.debug:
            print("\n\n#######################################################")
            print('CfnParser - parse'+lineno())
            print('Beginning to parse cloudformation template')
            print('cloudformation type: '+str(type(cloudformation_yml)))
            print("##########################################################\n\n")
        try:
            self.pre_validate_model(cloudformation_yml)

            if self.debug:
                print('successful pre_validate_model: '+lineno())

        except ParserError as e:
            tb = sys.exc_info()[-1]
            if self.debug:
                print('tb: ' + str(tb) + lineno())
            stk = traceback.extract_tb(tb, 1)
            if self.debug:
                print('stk: ' + str(stk) + lineno())
            fname = stk[0][2]
            if self.debug:
                print('The failing function was', fname,lineno())
            raise

        if self.debug:
            print("\n\n#########################################")
            print('cloudformation template pre_validated'+lineno())
            print('Prevalidating cloudformation template')
            print("#############################################\n\n")


        # Transform raw resources in template as performed by
        # transforms
        #transformer = TransformRegistry.TransformRegistry(debug=self.debug)
        #cloudformation_yml = transformer.perform_transforms(cloudformation_yml)

        #if self.debug:
        #    print("\n\n#################################################")
        #    print('Done prevalidating cloudformation template')
        #    print('cloudformation_yml type: '+str(type(cloudformation_yml))+lineno())
        #    print('Begin to validate referenses'+lineno())
        #    print("#####################################################\n\n")


        # Not validating references, because I believe terraform will do this internally
        #self.validate_references(cloudformation_yml)


        if self.debug:
            print("\n\n##########################################")
            print('Begin transform_hash_into_model elements'+lineno())
            print('Creating the cfn_model objects')
            print("#############################################\n")

        cfn_model = CfnModel.CfnModel(debug=self.debug)
        cfn_model.raw_model=cloudformation_yml


        # pass 1: wire properties into ModelElement objects
        cfn_model = self.transform_hash_into_model_elements(cloudformation_yml, cfn_model)

        if self.debug:
            print("\n\n#################################################")
            print("Iterate through each resource in the model"+lineno())
            print("######################################################\n")

            for r in cfn_model.resources:
                print("############### RESOURCE INFO ###################")
                print(r)
                print('resource_type: '+str(cfn_model.resources[r].resource_type)+lineno())
                print('logical_resource_id: '+str(cfn_model.resources[r].logical_resource_id)+lineno())
                print('metadata: '+str(cfn_model.resources[r].metadata)+lineno())
                print('vars: '+str(vars(cfn_model.resources[r]))+lineno())
                print('raw model: '+str(cfn_model.resources[r].raw_model.raw_model)+lineno())
                print("##################################################\n")

            print("\n##################################################")
            print('properties wired into model element objects'+lineno())
            print("##################################################\n")
            input('press enter to continue '+lineno())

        if self.debug:
            print("\n##################################################")
            print('Transforming hash into parameters'+lineno())
            print("##################################################\n")
            input('press enter to continue '+lineno())

        # Transform cloudformation parameters into parameters object
        cfn_model = self.transform_hash_into_parameters(cloudformation_yml,cfn_model)

        if self.debug:
            print("\n##################################################")
            print('Done transforming hash into parameters' + lineno())
            print('parameters: '+str(cfn_model.parameters)+lineno())
            print('resources: '+str(cfn_model.resources)+lineno())
            print('providers: '+str(cfn_model.providers)+lineno())
            print('locals: '+str(cfn_model.locals)+lineno())
            print('outputs: '+str(cfn_model.outputs)+lineno())
            print('data: '+str(cfn_model.data)+lineno())
            print('raw model: '+str(cfn_model.raw_model)+lineno())
            print('Beginning post process resource model elements' + lineno())
            print("##################################################\n")


            for r in cfn_model.resources:
                print("############### RESOURCE INFO ###################")
                print(r)
                print('resource_type: '+str(cfn_model.resources[r].resource_type)+lineno())
                print('logical_resource_id: '+str(cfn_model.resources[r].logical_resource_id)+lineno())
                print('metadata: '+str(cfn_model.resources[r].metadata)+lineno())
                print('vars: '+str(vars(cfn_model.resources[r]))+lineno())
                print('raw model: '+str(cfn_model.resources[r].raw_model.raw_model)+lineno())
                print("##################################################\n")

                print('vars: '+str(vars(cfn_model))+lineno())

            print('outputs: '+str(cfn_model.outputs)+lineno())
            for output in cfn_model.outputs:
                print(str(output)+lineno())
                print(str(cfn_model.outputs[output])+lineno())
                print('id: '+str(cfn_model.outputs[output].id)+lineno())
                print('type: '+str(cfn_model.outputs[output].type)+lineno())




        # pass 2: tie together separate resources only where necessary to make life easier for rule logic
        cfn_model = self.post_process_resource_model_elements(cfn_model)

        if self.debug:
            print("\n##################################################")
            print('Done Post processing resource model elements' + lineno())
            print('Beginning to apply parameter values to model' + lineno())
            print("##################################################\n")

        cfn_model = self.apply_parameter_values(cfn_model, parameter_values_json)

        if self.debug:
            print("\n##################################################")
            print('Done applying parameter values to model' + lineno())
            print("##################################################\n")


        return cfn_model


    def apply_parameter_values(self, cfn_model, parameter_values_json):
        """
        ???
        :param cfn_model: 
        :param parameter_values_json: 
        :return: 
        """
        if self.debug:
            print('CfnParser - apply_parameter_values'+lineno())
            print('parameter_values_json: '+str(parameter_values_json)+lineno())

            if parameter_values_json:
                print('parameter values json: '+str(parameter_values_json)+lineno())

        return cfn_model

    def post_process_resource_model_elements(self, cfn_model):
        """
        Post process the resource model elements
        :param cfn_model: 
        :return: 
        """
        if self.debug:
            print("\n\n#######################################################")
            print('CfnParser - post_process_resource_model_elements'+lineno())
            print('parameters: '+str(cfn_model.parameters)+lineno())
            print('resources: '+str(cfn_model.resources)+lineno())
            print('raw model: '+str(cfn_model.raw_model)+lineno())
            print("##########################################################\n\n")

        resource_parser_class = ParserRegistry(debug=self.debug)

        resource_map = {
            'aws_elb':'AWS::ElasticLoadBalancing::LoadBalancer',
            'aws_security_group':'AWS::EC2::SecurityGroup',
            'egress':'AWS::EC2::SecurityGroupEgress',
            'ingress':'AWS::EC2::SecurityGroupIngress',
            'aws_instance':'AWS::EC2::Instance',
            'aws_iam_role':'AWS::IAM::Role',
            'aws_iam_policy': 'AWS::IAM::Policy',
            'aws_iam_user': 'AWS::IAM::User',
            'aws_network_interface': 'AWS::EC2::NetworkInterface',
            'aws_s3_bucket_policy': 'AWS::S3::BucketPolicy',
            'aws_sqs_queue_policy': 'AWS::SQS::QueuePolicy',
            'aws_iam_policy_attachment': 'AWS::IAM::ManagedPolicy'
        }


        if self.debug:
            print("\n##############################")
            print("Iterating through each of the resources in the cfn model")
            print("################################\n")

        for r in cfn_model.resources:

            if cfn_model.resources[r].resource_type not in resource_map:
                #print('fix cfnparser model map: '+lineno())
                #sys.exit(1)
                return cfn_model

            if self.debug:
                print("\n\n#######################################")
                print("Processing cfn_model.resource: "+str(r)+lineno())
                print("########################################\n")

                print('resource type: '+str(type(cfn_model.resources[r].resource_type))+lineno())
                print('cfn_model type: '+str(cfn_model.resources[r].resource_type)+lineno())

                #print('parser registry:'+str(resource_parser_class.registry)+lineno())


            # if there is a resource parser in the registry
            if resource_map[cfn_model.resources[r].resource_type] in resource_parser_class.registry:

                if self.debug:
                    print('found it the resource parser we were looking for '+lineno())
                    print('type: '+str(cfn_model.resources[r].resource_type)+lineno())
                    #print('class: '+str(resource_parser_class.registry[cfn_model.resources[r].resource_type])+lineno())

                    print("\n\n############################")
                    print('Parsing resource: '+str(r)+lineno())
                    print('type: '+str(resource_parser_class.registry[resource_map[cfn_model.resources[r].resource_type]])+lineno())
                    print("################################\n")

                resource_parser = resource_parser_class.registry[resource_map[cfn_model.resources[r].resource_type]]

                if self.debug:
                    print('sending resource to parser: '+str(cfn_model.resources[r])+lineno())


                resource_parser.parse(cfn_model, cfn_model.resources[r], debug=self.debug)


        if self.debug:
            print('done parsing the cfn model')

        return cfn_model


    def transform_hash_into_model_elements(self, cfn_hash, cfn_model):
        """
        We are iterating the the resources in the cloudformation template
        and trying to create objects out of each resource, and then putting
        the objects in to the model object
        :param cfn_hash: 
        :param cfn_model: 
        :return: 
        """
        if self.debug:
            print("\n#############################################################")
            print('CfnParser - transform_hash_into_model_elements'+lineno())
            print('hash: '+str(cfn_hash)+lineno())
            print('cfn_model: '+str(cfn_model)+lineno())
            print("##############################################################\n")

        if type(cfn_hash) == type(str()):
            json_acceptable_string = cfn_hash.replace("'", "\"")
            cfn_hash= json.loads(json_acceptable_string)

        # Iterate through each of the resources
        for resource_type in cfn_hash['resource']:

            if self.debug:
                print("\n##############################################")
                print('resource_type: '+str(resource_type)+lineno())
                print("interating over each of the resources in the resource type")
                print("################################################\n")


            for resource_name in cfn_hash['resource'][resource_type]:

                if self.debug:
                    print("\n###############################################")
                    print('resource type: '+str(resource_type)+lineno())
                    print('resource name: '+str(resource_name)+lineno())
                    print("#################################################\n")

                # Create a new resource class based on the name of the resource
                # Does this by getting the resource type, and creating object based on resource type
                resource_class = self.class_from_type_name(resource_type,resource_name)

                if self.debug:
                    print("\n###########################################")
                    print('resource class: '+str(resource_class)+lineno())
                    print("############################################\n")
                # This is the original model
                resource_class.raw_model= cfn_model

                if self.debug:
                    print("\n##########################################")
                    print('Setting logical resource id to: '+str(resource_name)+lineno())
                    print("############################################")
                resource_class.logical_resource_id = resource_name

                resource_class.resource_type = resource_type
                if 'metadata' in cfn_hash['resource'][resource_type][resource_name]:
                    resource_class.metadata = cfn_hash['resource'][resource_type][resource_name]['metadata']

                if self.debug:
                    print("\n#############################################")
                    print("Assigning fields based upon properties"+lineno())
                    print("###############################################\n")

                resource_class = self.assign_fields_based_upon_properties(resource_class, cfn_hash['resource'][resource_type][resource_name])

                if self.debug:
                    print('There fields are no assigned to properties'+lineno())

                cfn_model.resources[resource_name]=resource_class

                if self.debug:
                    print("\n\n################################################")
                    print('The new resource object which is being added to cfn_model.resources has the following properties'+lineno())
                    print(type(resource_class))
                    print(vars(resource_class))
                    print(dir(resource_class))
                    print('resource_type: '+str(resource_class.resource_type)+lineno())
                    print('logical_resource_id: '+str(resource_class.logical_resource_id)+lineno())
                    print("##############################################################\n")

        if self.debug:
            print("\n#########################################")
            print("Model is")
            print('cfn model outputs:'+str(cfn_model.outputs)+lineno())
            print('cfn model locals:'+str(cfn_model.locals)+lineno())
            print('cfn model providers:'+str(cfn_model.providers)+lineno())
            print('cfn model parameters: '+str(cfn_model.parameters)+lineno())
            print('cfn model data:'+str(cfn_model.data)+lineno())
            print('cfn model resources:' +str(cfn_model.resources)+lineno())
            print('cfn model raw_model:'+str(cfn_model.raw_model)+lineno())

            input('check model - press enter to continue: '+lineno())

        return cfn_model

    def transform_hash_into_parameters(self, cfn_hash, cfn_model):
        """
        Transform hash into parameters
        :param cfn_hash: 
        :param cfn_model: 
        :return: 
        """

        if self.debug:
            print('CfnParser - transform_hash_into_parameters'+lineno())
            print('cfn_hash: '+str(cfn_hash)+lineno())
            print('type hash: '+str(type(cfn_hash))+lineno())
            print('cfn_model- raw model: '+str(cfn_model.raw_model)+lineno())


        if type(cfn_hash) == type(str()):
            json_acceptable_string = cfn_hash.replace("'", "\"")
            cfn_hash= json.loads(json_acceptable_string)

        if 'variable' in cfn_hash and cfn_hash['variable']:

            if self.debug:
                print('variable in hash: '+str(cfn_hash['variable'])+lineno())

            for variable in cfn_hash['variable']:

                parameter = Parameter(debug=self.debug)
                parameter.id = variable

                if self.debug:
                    print('parameter id: ' + str(parameter.id) + lineno())

                if cfn_hash['variable'][variable]:

                    if self.debug:
                        print('variable is: '+str(cfn_hash['variable'][variable])+lineno())
                        print("\n#####################################")
                        print("Iterating through each of the parameters in variable")
                        print("#######################################\n")

                    for param in cfn_hash['variable'][variable]:

                        if self.debug:
                            print("\n###############################")
                            print('param: '+str(param)+lineno())
                            print('value: '+str(cfn_hash['variable'][variable][param])+lineno())
                            print("##################################\n")

                        if param == 'default':
                            parameter.type = str(type(cfn_hash['variable'][variable][param]))
                        else:
                            if self.debug:
                                print('no default parameter'+lineno())

                        if self.debug:
                            print('adding parameter: '+str(variable)+'='+str(cfn_hash['variable'][variable][param])+lineno())

                        parameter.instance_variables.append(param+'='+str(cfn_hash['variable'][variable][param]))

                    cfn_model.parameters[variable] = parameter

                if self.debug:
                    print('cfn model parameters:'+str(cfn_model.parameters)+lineno())
                    input('wiring variable: '+str(variable)+lineno())

        if 'data' in cfn_hash and cfn_hash['data']:

            if self.debug:
                print('data in hash: '+str(cfn_hash['data'])+lineno())

            for cfdata in cfn_hash['data']:

                data = Data(debug=self.debug)
                data.id = cfdata

                if self.debug:
                    print('data is: '+str(cfdata)+lineno())
                    print('data id: ' + str(data.id) + lineno())

                if cfn_hash['data'][cfdata]:

                    if self.debug:
                        print('data is: '+str(cfn_hash['data'][cfdata])+lineno())
                        print("\n#####################################")
                        print("Iterating through each of the parameters in data")
                        print("#######################################\n")

                    for param in cfn_hash['data'][cfdata]:

                        if self.debug:
                            print("\n###############################")
                            print('data: '+str(param)+lineno())
                            print('param: '+str(param)+lineno())
                            print('value: '+str(cfn_hash['data'][cfdata][param])+lineno())
                            print("##################################\n")

                        if param == 'default':
                            data.type = str(type(cfn_hash['data'][cfdata][param]))
                        else:
                            if self.debug:
                                print('no default data'+lineno())

                        if self.debug:
                            print('adding data: '+str(cfdata)+'='+str(cfn_hash['data'][cfdata][param])+lineno())

                        data.instance_data.append(param+'='+str(cfn_hash['data'][cfdata][param]))

                    cfn_model.data[cfdata] = cfdata



        if 'local' in cfn_hash and cfn_hash['local']:

            if self.debug:
                print('locals in hash: '+str(cfn_hash['local'])+lineno())

            for cflocal in cfn_hash['local']:

                locals = Locals(debug=self.debug)
                locals.id = cflocal

                if self.debug:
                    print('locals id: ' + str(locals.id) + lineno())

                if cfn_hash['local'][cflocal]:

                    if self.debug:
                        print('local is: '+str(cfn_hash['local'][cflocal])+lineno())
                        print("\n#####################################")
                        print("Iterating through each of the parameters in locals")
                        print("#######################################\n")

                    for param in cfn_hash['local'][cflocal]:

                        if self.debug:
                            print("\n###############################")
                            print('local: '+str(param)+lineno())
                            print('value: '+str(cfn_hash['local'][cflocal][param])+lineno())
                            print("##################################\n")

                        if param == 'default':
                            locals.type = str(type(cfn_hash['local'][cflocal][param]))
                        else:
                            if self.debug:
                                print('no default local'+lineno())

                        if self.debug:
                            print('adding local: '+str(cflocal)+'='+str(cfn_hash['local'][cflocal][param])+lineno())

                        locals.instance_locals.append(param+'='+str(cfn_hash['local'][cflocal][param]))

                    cfn_model.locals[cflocal] = locals


        if 'output' in cfn_hash and cfn_hash['output']:

            if self.debug:
                print('outputs in hash: '+str(cfn_hash['output'])+lineno())

            for output in cfn_hash['output']:

                outputs = Output(debug=self.debug)
                outputs.id = output

                if self.debug:
                    print('outputs id: ' + str(outputs.id) + lineno())

                if cfn_hash['output'][output]:

                    if self.debug:
                        print('output is: '+str(cfn_hash['output'][output])+lineno())
                        print("\n#####################################")
                        print("Iterating through each of the parameters in outputs")
                        print("#######################################\n")



                    for param in cfn_hash['output'][output]:

                        if self.debug:
                            print("\n###############################")
                            print('local: '+str(param)+lineno())
                            print('value: '+str(cfn_hash['output'][output][param])+lineno())
                            print("##################################\n")

                        if param == 'default':
                            outputs.type = str(type(cfn_hash['output'][output][param]))
                        else:
                            if self.debug:
                                print('no default output'+lineno())

                        if self.debug:
                            print('adding output: '+str(output)+'='+str(cfn_hash['output'][output][param])+lineno())

                        outputs.instance_outputs.append(param+'='+str(cfn_hash['output'][output][param]))

                    cfn_model.outputs[output] = outputs



        if 'provider' in cfn_hash and cfn_hash['provider']:

            if self.debug:
                print('providers in hash: '+str(cfn_hash['provider'])+lineno())

            for provider in cfn_hash['provider']:

                providers = Provider(debug=self.debug)
                providers.id = provider

                if self.debug:
                    print('providers id: ' + str(providers.id) + lineno())

                if cfn_hash['provider'][provider]:

                    if self.debug:
                        print('provider is: '+str(cfn_hash['provider'][provider])+lineno())
                        print("\n#####################################")
                        print("Iterating through each of the parameters in providers")
                        print("#######################################\n")

                    for param in cfn_hash['provider'][provider]:

                        if self.debug:
                            print("\n###############################")
                            print('local: '+str(param)+lineno())
                            print('value: '+str(cfn_hash['provider'][provider][param])+lineno())
                            print("##################################\n")

                        if param == 'default':
                            providers.type = str(type(cfn_hash['provider'][provider][param]))
                        else:
                            if self.debug:
                                print('no default provider'+lineno())

                        if self.debug:
                            print('adding provider: '+str(provider)+'='+str(cfn_hash['provider'][provider][param])+lineno())

                        providers.instance_providers.append(param+'='+str(cfn_hash['provider'][provider][param]))

                    cfn_model.providers[provider] = providers

        return cfn_model


    def pre_validate_model(self, cloudformation_yml):
        """
        Prevalidate the cloudformation template
        :param cloudformation_yml: 
        :return: 
        """

        if self.debug:
            print("\n\n######################################")
            print('CfnParser - pre_validate_model'+lineno())
            print('########################################\n\n')
            print('cloudformation_yml: '+str(cloudformation_yml)+lineno())

        validator = CloudformationValidator.CloudformationValidator(debug=self.debug)

        errors = validator.validate(cloudformation_yml)

        if errors and len(errors)>0:

            try:

                raise ParserError('Basic CloudFormation syntax error', errors,debug=self.debug)
            except ParserError as e:
                tb = sys.exc_info()[-1]
                if self.debug:
                    print('tb: '+str(tb)+lineno())
                    print('to hash: '+str(e.to_hash())+lineno())
                stk = traceback.extract_tb(tb, 1)
                if self.debug:
                    print('stk: '+str(stk)+lineno())
                fname = stk[0][2]
                if self.debug:
                    print('The failing function was', fname,lineno())
                    print('parser error '+lineno())
                # FIXME
                raise ParserError('Basic CloudFormation syntax error', errors,debug=self.debug)


    def validate_references(self, cfn_hash):
        """
        Validate references in the cloudformation template
        :param cfn_hash: 
        :return: 
        """
        if self.debug:
            print("\n\n###########################################")
            print('CfnParser - validate_references'+lineno())
            print('Looking for unresolved references')
            print('cfn_hash len: '+str(len(cfn_hash))+lineno())
            print("################################################\\n\n")

        ref_validator = ReferenceValidator.ReferenceValidator(debug=self.debug)
        unresolved_refs = ref_validator.unresolved_references(cfn_hash)

        if unresolved_refs and len(unresolved_refs)>0:
            raise ParserError("Unresolved logical resource ids: "+str(unresolved_refs))

    def assign_fields_based_upon_properties(self, resource_object, resource):
        """
        Assign fields in the object based on properties in the resource
        :param resource_object: 
        :param resource: 
        :return: 
        """

        if self.debug:
            print("\n\n#######################################################")
            print('CfnParser - assign_fields_based_upon_properties'+lineno())
            print("#########################################################\n\n")

            print('resource_object: '+str(resource_object)+lineno())
            print('resource: '+str(resource)+lineno())
            print('resource_type:'+str(resource))

        resource_type = str(resource)

        if self.debug:
            print('resource_object dir: '+str(dir(resource_object))+lineno())
            print('resource_object vars: '+str(vars(resource_object))+lineno())


        for p in resource:

            if self.debug:
                print('property: '+str(p)+lineno())
            #new_property_name = self.map_property_name_to_attribute(p)

            if self.debug:
                print("\n\n###########################################")
                print('Creating object attribute during mapping'+lineno())
                print('property: '+str(p))
                #print('property details: '+str(resource['Properties'][p]))
                #print('dir:'+str(dir(resource_object))+lineno())
                #print('new property name'+str(new_property_name)+lineno())
                print('property name: '+str(p)+lineno())
                print('property value: '+str(resource[p])+lineno())
                print("###############################################\n")

            setattr(resource_object, p, str(resource[p]) )

        if self.debug:
            print("\n\n#############################################")
            print('The new resource object now has the following properties'+lineno())
            if hasattr(resource_object, 'policies'):
                print('policies: '+str(resource_object.policies))

            if hasattr(resource_object, 'policy_objects'):
                print('policy objects: '+str(resource_object.policy_objects))
            if hasattr(resource_object, 'group_names'):
                print('group names: '+str(resource_object.group_names))
            if hasattr(resource_object, 'groups'):
                print('groups: '+str(resource_object.groups))
            if hasattr(resource_object, 'resource_type'):
                print('resource type: '+str(resource_object.resource_type))
            if hasattr(resource_object, 'policy_document'):
                print('policy document: '+str(resource_object.policy_document))
            print(vars(resource_object))
            print("##################################################\n")

        if self.debug:
            print('resource_object dir: '+str(dir(resource_object))+lineno())
            print('resource_object vars: '+str(vars(resource_object))+lineno())



        return resource_object

    def map_property_name_to_attribute(self, string):
        """
        Mapp properties to attributes
        :param string: 
        :return: 
        """

        if self.debug:
            print('property name: '+str(string)+lineno())

        #first_character = str(string)[:1]

        #if self.debug:
        #    print('first character: '+str(first_character)+lineno())

        #first_character=first_character.lower()
        #remaining_characters = string[1:]

        #new_property_name = str(first_character)+str(remaining_characters)
        #new_property_name.replace('-','_')

        #if self.debug:
        #    print('new_property_name: '+str(new_property_name)+lineno())

        #return new_property_name

        return string

    def class_from_type_name(self, type_name, cfn_model):
        """
        Create class from type name
        :param type_name: 
        :param cfn_model: 
        :return: 
        """

        if self.debug:
            print('CfnParser - class_from_type_name'+lineno())
            print('type_name: '+str(type_name)+lineno())

        resource_class = self.generate_resource_class_from_type(type_name,cfn_model)

        return resource_class

    def generate_resource_class_from_type(self, type_name, cfn_model):
        """
        Generate resource class from type
        :param type_name: 
        :param cfn_model: 
        :return: 
        """

        if self.debug:
            print("\n##################################################")
            print('Getting the resource class')
            print('generate_resource_class_from_type'+lineno())
            print('cfn model: '+str(cfn_model)+lineno())
            print('type name: '+str(type_name)+lineno())
            print("###################################################\n")


        models = [
            'S3BucketPolicy',
            'EC2Instance',
            'EC2NetworkInterface',
            'IAMGroup',
            'IAMManagedPolicy',
            'IAMPolicy',
            'IAMRole',
            'IAMUser',
            'ElasticLoadBalancingLoadBalancer',
            'ElasticLoadBalancingV2LoadBalancer',
            'SQSQueuePolicy',
            'EC2SecurityGroup',
            'EC2SecurityGroupEgress',
            'EC2SecurityGroupIngress',
            'SNSTopicPolicy'
        ]

        models2= [
            'aws_instance',
            'aws_elb',
            'aws_security_group',
            'aws_security_group_rule',
            'aws_iam_role',
            'aws_iam_policy',
            'aws_iam_user',
            'aws_network_interface',
            'egress',
            'ingress',
            'aws_iam_group',
            'aws_s3_bucket_policy'
            'aws_sqs_queue_policy',
            'aws_iam_policy_attachment'
        ]
        module_name = 'AWS'

        if type_name == 'aws_iam_policy_attachment':
            short_name == 'IAMManagedPolicy'
        if type_name == 'aws_sqs_queue_policy':
            short_name == 'SQSQueuePolicy'
        if type_name == 'aws_s3_bucket_policy':
            short_name == 'S3BucketPolicy'
        if type_name == 'aws_iam_group':
            short_name == 'IAMGroup'
        if type_name == 'egress':
            short_name == 'EC2SecurityGroupEgress'
        if type_name == 'ingress':
            short_name == 'EC2SecurityGroupIngress'
        if type_name == 'aws_network_interface':
            short_name == 'EC2NetworkInterface'
        if type_name == 'aws_iam_user':
            short_name = 'IAMUser'
        if type_name == 'aws_iam_policy':
            short_name = 'IAMPolicy'
        if type_name == 'aws_instance':
            short_name ='EC2Instance'
        if type_name == 'aws_elb':
            short_name = 'ElasticLoadBalancingLoadBalancer'
        if type_name == 'aws_security_group':
            short_name = 'EC2SecurityGroup'
        if type_name == 'aws_iam_role':
            short_name = 'IAMRole'
        if type_name == 'aws_security_group_rule':
            if type == 'x':
                short_name = 'EC2SecurityGroupEgress'
            else:
                short_name = 'EC2SecurityGroupIngress'

        #module_names = type_name.split('::')
        if module_name == 'Custom':

            if self.debug:
                print('module is custom: '+lineno())


            custom_resource_class_name = self.initial_upper(short_name)
            resource_class = ModelElement.ModelElement(debug=self.debug)
            setattr(resource_class,'__name__',custom_resource_class_name)
            setattr(resource_class,'debug',self.debug)
        elif module_name == 'AWS':

            if self.debug:
                print('module is AWS: '+lineno())


            if str(type_name) in models2:

                if self.debug:
                    print('short_name: ' + str(short_name) + lineno())
                    print('sys modules __name__: '+str(sys.modules[__name__])+lineno())

                resource_class = getattr(sys.modules[__name__], short_name).__getattribute__(short_name)(cfn_model)
                setattr(resource_class,'debug',self.debug)

                if self.debug:
                    print('vars: ' + str(vars(resource_class)) + lineno())




            else:

                if self.debug:
                    print('type: '+str(type_name)+' not a resource we are concered with.  Making generic model')

                resource_class = ModelElement.ModelElement(cfn_model)
                setattr(resource_class, '__name__', type_name)
                setattr(resource_class,'debug',self.debug)
        else:
            print('unknown namespace in resource type'+lineno())
            sys.exit(1)



        return resource_class


    def initial_upper(self, string):
        """
        First character to upper case
        :param string: 
        :return: 
        """
        if self.debug:
            print('CfnParser - initial_upper'+lineno())

        first_character = str(string)[:1]
        remaining_characters = string[1:]

        new_property_name = str(first_character.upper())+str(remaining_characters)

        return new_property_name


from __future__ import absolute_import, division, print_function
import inspect
import sys
from terraform_model.model.Principal import Principal

def lineno():
    """Returns the current line number in our program."""
    return str(' - Statement - line number: '+str(inspect.currentframe().f_back.f_lineno))

class Statement:
    """
    Statement model
    """
    def __init__(self, debug=False):
        """
        Initialize
        :param debug: 
        """
        self.debug=debug

        if self.debug:
            print("\n\n#############################")
            print('Statement __init__'+lineno())
            print('debug: '+str(debug)+lineno())
            print("################################\n")

        self.actions=[]
        self.not_actions = []
        self.resources = []
        self.not_resources = []
        self.sid = None
        self.effect = None
        self.condition = None
        self.principal = None
        self.not_principal = None

        if self.debug:
            print('init')

    def wildcard_actions(self):
        """
        wildcard actions
        :return: 
        """
        if self.debug:
            print('wildcard_actions'+lineno())

        actions= []

        for action in self.actions:
            if self.debug:
                print('action: '+str(action)+lineno())
                print('type: '+str(type(action))+lineno())

            if type(action)== type(str()):
                if '*' in action:
                    actions.append(action)

            if sys.version_info[0] < 3:

                if type(action) == type(unicode()):
                    if '*' in action:
                        actions.append(action)

            elif type(action)==type(list()):
                for item in action:


                    if hasattr(item,'__iter__'):
                        if type(item) == str():
                            if self.debug:
                                print('is iterable'+lineno())
                            if '*' in item:
                                actions.append(item)
                    else:
                        if self.debug:
                            print("No match!!"+lineno())

        return actions

    def wildcard_principal(self):
        """
        Get wildcard principals
        :return: 
        """
        if self.debug:
            print('wildcard_principal'+lineno())

        principal = Principal(debug=self.debug)
        return principal.wildcard(self.principal)

    def wildcard_resources(self):
        """
        Get wildcard resources
        :return: 
        """
        if self.debug:
            print('wildcard_resources'+lineno())

        resources = []

        for resource in self.resources:
            if self.debug:
                print('resource: '+str(resource)+lineno())
                print('type: '+str(type(resource))+lineno())

            if type(resource) == None:
                if self.debug:
                    print('is none: '+lineno())
                continue
            if type(resource)== type(str()):
                if self.debug:
                    print('is string '+lineno())
                if '*' in resource:
                    resources.append(resource)

            if sys.version_info[0] < 3:
                if self.debug:
                    print('is python2 '+lineno())

                if type(resource) == type(unicode()):
                    if '*' in resource:
                        resources.append(resource)


            if type(resource)== type(dict()):
                if self.debug:
                    print('is a dict'+lineno())
                for item in resource:
                    if self.debug:
                        print('item: '+str(item)+lineno())
                        print('resource item: '+str(resource[item])+lineno())

                    if type(resource[item]) == type(list()):
                        for more_items in resource[item]:
                            if self.debug:
                                print('more items: '+str(more_items)+lineno())

                            if type(more_items) == type(str()):
                                if '*' in more_items:
                                    resources.append(more_items)

                            if sys.version_info[0] < 3:
                                if type(more_items) == type(unicode()):
                                    if '*' in more_items:
                                        resources.append(more_items)

                            elif type(more_items) == type(list()):
                                if self.debug:
                                    print('is a list'+lineno())
                                for many_more_items in more_items:
                                    if self.debug:
                                        print('many more items: '+str(many_more_items)+lineno())

                                    if type(many_more_items)== type(str()):
                                        if '*' in many_more_items:
                                            resources.append(many_more_items)

                                    if sys.version_info[0] < 3:
                                        if type(many_more_items) == type(unicode()):
                                            if '*' in many_more_items:
                                                resources.append(many_more_items)

            if type(resource) == type(list()):
                for item in resource:
                    if self.debug:
                        print('item: '+str(item)+lineno())

                    try:
                        some_object_iterator = iter(item)
                        if type(item) == str():
                            if self.debug:
                                print('is iterable' + lineno())
                            if '*' in item:
                                resources.append(item)

                    except TypeError as te:
                        if self.debug:
                            print("No match!!"+lineno())

        return resources


from __future__ import absolute_import, division, print_function
from terraform_model.model.ModelElement import ModelElement


class EC2SecurityGroupEgress(ModelElement):
    """
    Ec2 security group egress model
    """
    def __init__(self, cfn_model, debug=False):
        """
        Initialize
        :param cfn_model:
        :param debug:
        """
        self.debug = debug

        if self.debug:
            print('EC2SecurityGroupEgress __init__')

        ModelElement.__init__(self, cfn_model)

        self.resource_type = 'AWS::EC2::SecurityGroupEgress'






from __future__ import absolute_import, division, print_function
import sys
import inspect
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule


def lineno():
    """Returns the current line number in our program."""
    return str(' -  IamManagedPolicyWildcardActionRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class IamManagedPolicyWildcardActionRule(BaseRule):
  
  def __init__(self, cfn_model=None, debug=None):
    """
    Initialize IamManagedPolicyWildcardActionRule
    :param cfn_model:
    """
    BaseRule.__init__(self, cfn_model, debug=debug)

  def rule_text(self):
    """
    Return rule text
    :return:
    """
    if self.debug:
        print('rule_text'+lineno())
    return 'IAM managed policy should not allow * action'
  

  def rule_type(self):
    """
    Return rule type
    :return:
    """
    self.type= 'VIOLATION::FAILING_VIOLATION'
    return 'VIOLATION::FAILING_VIOLATION'
  

  def rule_id(self):
    """
    Return rule id
    :return:
    """
    if self.debug:
        print('rule_id'+lineno())
    self.id ='F5'
    return 'F5'
  

  def audit_impl(self):
    """
    Audit
    :return: violations
    """
    if self.debug:
        print('IamManagedPolicyWildcardActionRule - audit_impl'+lineno())
    
    violating_policies = []

    resources = self.cfn_model.resources_by_type('AWS::IAM::ManagedPolicy')

    if len(resources)>0:
      for resource in resources:
          if self.debug:
              print('resources: '+str(resource)+lineno())
              print('vars: '+str(vars(resource))+lineno())
          if hasattr(resource,'policy_document'):
            if self.debug:
                print('has policy document '+lineno())
                print('vars: '+str(vars(resource.policy_document)))

            if resource.policy_document:
                if self.debug:
                    print(resource.policy_document.statements)

                for statements in resource.policy_document.statements:
                    if self.debug:
                        print('statements: '+str(statements))
                        print('vars: '+str(vars(statements)))

                if resource.policy_document.wildcard_allowed_actions():
                    if self.debug:
                        print('has wildcard allow actions????')

                    violating_policies.append(str(resource.logical_resource_id))

          else:
            if self.debug:
                print('does not have policy document'+lineno())

    else:
      if self.debug:
        print('no violating_policies' + lineno())

    if self.debug:
      input('Press enter to continue: '+lineno())

    return violating_policies


from __future__ import absolute_import, division, print_function
import sys
import inspect
from pykwalify.core import Core
from terraform_model.validator import SchemaGenerator
from schema import Schema, And, Use, Optional, SchemaError

def lineno():
    """Returns the current line number in our program."""
    return str(' -  CloudformationValidator- line number: '+str(inspect.currentframe().f_back.f_lineno))

class CloudformationValidator:
    """
    Cloudformation validator
    """
    def __init__(self, debug=False):
        """
        Initialize
        :param debug:
        """

        self.debug = debug

        if self.debug:
            print('CloudformationValidator - __init__'+lineno())

    def validate(self, cloudformation_string):
        """
        Validating the schema for a cloudformation segment
        :param cloudformation_string:
        :return:
        """
        if self.debug:
            print("\n\n##################################################")
            print('CloudformationValidator - validate - validating following string'+lineno())

            print(str(cloudformation_string)+lineno())
            print("#######################################################\n\n")

        #schema = SchemaGenerator.SchemaGenerator(debug=self.debug)

        #if self.debug:
        #    print('schema: '+str(schema)+lineno())

        #main_schema = schema.generate(cloudformation_string)

        #if self.debug:
        #    print('main_schema: '+str(main_schema)+lineno())

        # FIXME
        ## create validator
        #c = Core(source_data=cloudformation_string, schema_data=main_schema)
        #c.validate(raise_exception=False)

        #if len(c.errors)>0:
        #    if self.debug:
        #        print('errors: '+str(c.errors)+lineno())
            #raise ParserError.new('Basic CloudFormation syntax error', errors)

        # Return any validation error
        #return c.errors


        if self.debug:
            print("\n##########################")
            print('Done with cloudformation validation'+lineno())
            print("############################\n")

        return None


    def check(self, conf_schema, conf):
        """
        ???
        :param conf_schema:
        :param conf:
        :return:
        """

        try:
            conf_schema.validate(conf)
            return True
        except SchemaError:
            return False

        conf_schema = Schema({
            'version': And(Use(int)),
            'info': {
                'conf_one': And(Use(float)),
                'conf_two': And(Use(str)),
                'conf_three': And(Use(bool)),
                Optional('optional_conf'): And(Use(str))
            }
        })

        conf = {
            'version': 1,
            'info': {
                'conf_one': 2.5,
                'conf_two': 'foo',
                'conf_three': False,
                'optional_conf': 'bar'
            }
        }

        if self.debug:
            print(check(conf_schema, conf))

            print('validate')


from __future__ import absolute_import, division, print_function
from terraform_model.model.ModelElement import ModelElement



class EC2Instance(ModelElement):
    """
    Ecs instance model
    """
    def __init__(self, cfn_model):
        """
        Initialize
        :param cfn_model: 
        """
        # attr_accessor :security_groups
        ModelElement.__init__(self, cfn_model)

        self.securityGroupIds= []
        self.networkInterfaces= []
        self.security_groups= []
        self.resource_type = 'AWS::EC2::Instance'






from __future__ import absolute_import, division, print_function
import logging
import json
import sys
import os
import traceback
import inspect
import yaml
import hcl
import re
import tempfile
import shutil
from terraform_model.parser import CfnParser
from terraform_validator.CustomRuleLoader import CustomRuleLoader
from terraform_validator.ProfileLoader import ProfileLoader
from terraform_validator.TemplateDiscovery import TemplateDiscovery
from terraform_validator.Violation import Violation
from terraform_validator.result_views.JsonResults import JsonResults
from terraform_validator.result_views.SimpleStdoutResults import SimpleStdoutResults
from terraform_model.parser.ParserError import ParserError

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

def lineno():
    """Returns the current line number in our program."""
    return str(' - ValidateUtility - line number: '+str(inspect.currentframe().f_back.f_lineno))

class GenericScalar:
    def __init__(self, value, tag, style=None):
        self._value = value
        self._tag = tag
        self._style = style

    @staticmethod
    def to_yaml(dumper, data):
        # data is a GenericScalar
        return dumper.represent_scalar(data._tag, data._value, style=data._style)


def default_constructor(loader, tag_suffix, node):
    if isinstance(node, yaml.ScalarNode):
        return GenericScalar(node.value, tag_suffix, style=node.style)
    else:
        raise NotImplementedError('Node: ' + str(type(node)))



class ValidateUtility:


    _template = None

    def __init__(self, config_block):
        """
        Initialize ValidateUtility
        :param config_block: 
        """
        self.input_path = None
        self.template_file = None
        self.debug = False
        self.profile_path = None
        self.rules_directory = None
        self.optional_rules_directory = None
        self.allow_suppression = True
        self.print_suppression = False
        self.isolate_custom_rule_exceptions = False
        self.print_suppression = True
        self.parameter_values_path = None
        self.isolate_custom_rule_exceptions = True
        self.suppress_errors = False
        self.use_optional_rules = False
        self.excluded_rules = []
        self.s3_bucket_name = None
        self.s3_profile = None




        if config_block:
            self._config = config_block
        else:
            logging.error('config block was garbage')
            raise SystemError

        for key in self._config:
            self.__dict__[key] = self._config[key]

        if self.use_optional_rules:
            self.optional_rules_directory = os.path.dirname(os.path.realpath(__file__))+'/additional_custom_rules'

        if self.s3_bucket_name:
            self.temp_dir_path = tempfile.mkdtemp()
        else:
            self.temp_dir_path = None

        if self.debug:
            print('s3 bucket: '+str(self.s3_bucket_name))
            print('temp directory: '+str(self.temp_dir_path))



        self.custom_rule_loader = CustomRuleLoader(
            isolate_custom_rule_exceptions=self.isolate_custom_rule_exceptions,
            allow_suppression=self.allow_suppression,
            print_suppression=self.print_suppression,
            debug=self.debug,
            additional_rules_directory = self.optional_rules_directory,
            excluded_rules=self.excluded_rules,
            s3_bucket_name = self.s3_bucket_name,
            temp_dir_path = self.temp_dir_path,
            s3_profile = self.s3_profile)

        # If we have an extra rules directory
        if self.rules_directory:
            self.custom_rule_loader.extra_rule_directory = self.rules_directory


        self.custom_rule_loader.rule_directory=str(os.path.dirname(os.path.realpath(__file__))+'/custom_rules')

        self.input_path = self.validate_path(self.input_path)
        self.template_file = self.validate_path(self.template_file)
        self.rules_directory = self.validate_path(self.rules_directory)
        self.profile_path = self.validate_path(self.profile_path)


    def validate(self):
        """
        Validate a directory or file
        :return: rendered results
        """

        if self.debug:
            print('ValidateUtility - validate'+lineno())

        input_path = self.input_path
        if self.debug:
            print('input_path: '+str(input_path)+lineno())


            print('\n\n############################################')
            print('audit_audit_aggregate_across_files_and_render_results '+lineno())
            print('################################################\n\n')


        aggregate_results = self.audit_aggregate_across_files_and_render_results(input_path=input_path)

        if self.debug:
            print('aggregate_results '+str(aggregate_results)+lineno())

        rendered_results =  self.render_results(aggregate_results=aggregate_results, output_format='json')


        if self.debug:
            print('rendered_results: '+str(rendered_results)+lineno())

        try:
            print(rendered_results)
            return rendered_results
        except Exception as e:
            return e


    def audit_aggregate_across_files_and_render_results(self,input_path=None, parameter_values_path=None, template_pattern=None):
        """
        Aggregate across files and render results
        :param input_path: 
        :param parameter_values_path: 
        :param template_pattern: 
        :return: 
        """
        if self.debug:
            print('ValidateUtility - audit_aggregate_across_files_and_render_results'+lineno())

        aggregate_results = self.audit_aggregate_across_files(input_path=input_path)

        if self.debug:
            print('results: '+str(aggregate_results)+lineno())

        return aggregate_results

    def merge_dicts(self, x, y):

        if type(x) == type(str()):
            json_acceptable_string = x.replace("'", "\"")
            x= json.loads(json_acceptable_string)


        if type(y) == type(str()):
            json_acceptable_string = y.replace("'", "\"")
            y= json.loads(json_acceptable_string)


        if self.debug:
            print('x: '+str(x)+lineno())
            print('y:'+str(y)+lineno())

        z = x.copy()  # start with x's keys and values
        z.update(y)  # modifies z with y's keys and values & returns None

        if self.debug:
            print('z: '+str(z)+lineno())

        return z


    def audit_aggregate_across_files(self,input_path=None, parameter_values_path=None, template_pattern=None):
        """
        Given a file or directory path, return aggregate results
        :param input_path: 
        :param parameter_values_path: 
        :param template_pattern: 
        :return: 
        """
        if self.debug:
            print('ValidateUtility - audit_aggregate_across_files'+lineno())
            print('input_path: '+str(input_path)+lineno())

        if parameter_values_path:
            with open(parameter_values_path,'r') as myparamsfile:
                param_data = myparamsfile.read().replace('\n','')

            parameter_values_string = param_data
        else:
            parameter_values_string = None

        if not self.template_file:
            templates_obj = TemplateDiscovery(input_json_path=input_path,debug=self.debug)
            templates = templates_obj.discover_templates()
        else:
            templates=[self.template_file]

        if self.debug:
            print(templates)

        aggregate_results = []



        file_data = {
            'variable':{},
            'resource':{},
            'data':{},
            'output':{},
            'locals':{},
            'provider':{}

        }


        for template in templates:

            if self.debug:
                print('template: '+str(template)+lineno())
                print('type: '+str(type(template))+lineno())


            if not self.suppress_errors:
                print("\nEvaluating: "+str(template)+lineno())

            json_matchObj = re.match(r'.*.json$', template, re.M | re.I)
            tf_matchObj = re.match(r'.*.tf$',template, re.M | re.I)


            if self.debug:
                print('json matchobj: '+str(json_matchObj)+lineno())
                print('tf matchobj: '+str(tf_matchObj)+lineno())

            if json_matchObj:

                with open(template, 'r') as myfile:
                    if self.debug:
                        print("\n"+'Auditing file: '+str(template)+lineno())

                    json_acceptable_string = myfile.read().strip("'<>() ")
                    json_acceptable_string=json_acceptable_string.replace("'","\'")

                    if self.debug:
                        print(json_acceptable_string)

                    try:
                        file_data = file_data + str(json.loads(json_acceptable_string))
                    except Exception as e:

                        if not self.suppress_errors:
                            print("\n##############################")
                            print("Invalid json file - "+str(template)+' - skipping file')
                            print("################################\n")

            elif tf_matchObj:

                #tf.add_multi_constructor('', default_constructor, Loader=yaml.SafeLoader)
                #tf.add_representer(GenericScalar, GenericScalar.to_yaml, Dumper=yaml.SafeDumper)

                with open(template, 'r') as myfile:

                    if self.debug:
                        print("\n#######################################")
                        print('template read: '+str(template)+lineno())
                        print("######################################\n")
                    try:

                        data = hcl.load(myfile)
                        #print('tempdata: '+str(temp_data))
                        #data = str(json.dumps(temp_data)).strip()

                        file_data = self.merge_dicts(file_data,data)


                        if self.debug:
                            print('data is: '+str(data)+lineno())


                    except Exception as e:
                        print("\n##############################")
                        print("Invalid tf file - " + str(template) + ' - skipping file')
                        print('error: '+str(e))
                        print("################################\n")

                    #file_results = self.audit(cloudformation_string=file_data,
                    #                          parameter_values_string=parameter_values_string)

                    #if self.debug:
                    #    print('file results: ' + str(file_results) + lineno())

                    #results = {}

                    #if 'failure_count' in file_results:
                    #    results['failure_count'] = file_results['failure_count']
                    #else:
                    #    results['failure_count'] = 1
                    #results['filename'] = template
                    #results['file_results'] = file_results
                    #aggregate_results.append(results)

            else:

                if self.debug:
                    print('not sure what this is: '+lineno())
                #try:
                #    json_stuff = open(template)
                #    my_template = json.load(json_stuff)

                #    if my_template and ('resources' in my_template or 'variable' in my_template):
                #        if self.debug:
                #            print('might be terraform '+lineno())

                #        with open(template, 'r') as myfile:
                #            if self.debug:
                #                print("\n" + 'Auditing file: ' + str(template) + lineno())

                #            json_acceptable_string = myfile.read().strip("'<>() ")
                #            json_acceptable_string = json_acceptable_string.replace("'", "\'")

                #            if self.debug:
                #                print(json_acceptable_string)

                #            try:
                #                data = json.loads(json_acceptable_string)
                #            except Exception as e:

                #                if not self.suppress_errors:
                #                    print("\n#################################")
                #                    print("Invalid json file - " + str(template) + ' - skipping file')
                #                    print("################################\n")

                #                data = {}
                #            file_results = self.audit(cloudformation_string=data,
                #                                      parameter_values_string=parameter_values_string)

                #            if self.debug:
                #                print('file results: ' + str(file_results) + lineno())

                #            results = {}

                #            if 'failure_count' in file_results:
                #                results['failure_count'] = file_results['failure_count']
                #            else:
                #                results['failure_count'] = 1
                #            results['filename'] = template
                #            results['file_results'] = file_results
                #            aggregate_results.append(results)

                #    else:
                #        if self.debug:
                #            print('file is not json template'+lineno())
                #except Exception as x:
                #    if self.debug:
                #        print('Exception caught in determining whether json template: '+str(x)+lineno())


        #if file_data.endswith(','):
        #    print('test3'+lineno())
        #    file_data = file_data[:-1]

        #file_data = file_data + '}'

        if self.debug:
            print('file data is now: ' + str(file_data) + lineno())

        #if type(file_data) == type(str()):
        #    json_acceptable_string = file_data.replace("'", "\"")
        #    file_data= json.loads(json_acceptable_string)


        #if self.debug:
        #    print('file data: '+str(file_data))
        #    print("\n###########################")
        #    print("Starting to audit template "+lineno())
        #    print("#############################\n")

        if self.debug:
            print('made it'+lineno())
            print('parameter values: '+str(parameter_values_string)+lineno())

        file_results = self.audit(cloudformation_string=file_data, parameter_values_string= parameter_values_string)


        if self.debug:
            print('file results: '+str(file_results)+lineno())

        results = {}

        if 'failure_count' in file_results:
            results['failure_count'] = file_results['failure_count']
        else:
            results['failure_count']= 1
        results['filename'] = template
        results['file_results'] = file_results
        aggregate_results.append(results)

        return aggregate_results

    def audit(self, cloudformation_string, parameter_values_string=None):
        """
        Audit files
        :param cloudformation_string: 
        :param parameter_values_string: 
        :return: 
        """

        if self.debug:
            print('ValidateUtility - audit'+lineno())

        violations = []

        parser = CfnParser.CfnParser(debug=self.debug)

        # We should validate the schema, just now sure how to do that right now, and don't
        # feel like spending any time on it
        schema_validation_errors = None

        if self.debug:
            print('schema validated: '+lineno())

        try:

            if self.debug:
                print("\n#########################")
                print("Starting to parse the template "+lineno())
                print('trying to parse cloudformation string: '+str(cloudformation_string)+lineno())
                print("###########################\n")

            cfn_model = parser.parse(cloudformation_string, parameter_values_string)

        except Exception as e:
            if self.debug:
                print('exception: '+str(e)+lineno())
                print('type: '+str(type(e))+lineno())
                if hasattr(e, 'to_hash'):
                    print('to_hash: '+str(e.to_hash())+lineno())
            cfn_model = None
            tb = sys.exc_info()[-1]
            if self.debug:
                print('tb: ' + str(tb) + lineno())
            stk = traceback.extract_tb(tb, 1)
            if self.debug:
                print('stk: ' + str(stk) + lineno())
            fname = stk[0][2]

            if not self.suppress_errors:
                print('The failing function was', fname, lineno())
                print('error:'+str(e)+lineno())

            if hasattr(e, 'to_hash'):
                schema_validation_errors=e.to_hash()
            else:
                schema_validation_errors=e
        if self.debug:
            if cfn_model:
                print('cfn_model: '+str(dir(cfn_model))+lineno())
                print('cfn_model: '+str(vars(cfn_model))+lineno())
                print(dir(cfn_model))

                for parameters in cfn_model.parameters:
                    print('parameter: '+str(parameters))
                    print(cfn_model.parameters[parameters])

                print("\n\n###############################")
                print('Done parsing parameters'+lineno())
                print("###############################\n\n")

                for resource in cfn_model.resources:
                    print('resource: '+str(resource))
                    print(cfn_model.resources[resource])
                    print('vars: '+str(vars(cfn_model.resources[resource])))
                    print('dir: '+str(dir(cfn_model.resources[resource])))
                    for attr in dir(cfn_model.resources[resource]):
                        print("obj.%s = %r" % (attr, getattr(cfn_model.resources[resource], attr)))

                print("\n\n###############################")
                print('Done parsing resources'+lineno())
                print("###############################\n\n")
                print('model is loaded'+lineno())

                print('vars: '+str(vars(cfn_model))+lineno())
                print('parameters: '+str(cfn_model.parameters))

        if not schema_validation_errors:
            if cfn_model:
                # Get any violations from custom rules
                # custom rules are in addition to the normal rules
                violations = self.custom_rule_loader.execute_custom_rules(cfn_model)

            if self.debug:
                print('violations: '+str(violations)+lineno())

                print("\n\n###############################")
                print('Done loading custom rules'+lineno())
                print("###############################\n\n")

            if len(violations)<1:
                failure_count = 0
            else:

                failure_count = self.count_failures(violations)
                if self.debug:
                    print('failure count: '+str(failure_count)+lineno())

                warning_count = self.count_warnings(violations)
                if self.debug:
                    print('warning count: '+str(warning_count))

            if self.debug:
                print('count: '+str(failure_count)+lineno())

            violations = self.filter_violations_by_profile(
                {
                    'failure_count': self.count_failures(violations),
                    'violations' : violations
                }
            )

            if self.debug:
                print('violations: '+str(violations)+lineno())
            return violations
        else:
            if self.debug:
                print('there are schema validation errors '+lineno())

            violation = Violation('FATAL','VIOLATION::FAILING_VIOLATION', str(schema_validation_errors))
            violations.append(violation)

            return violations

    def filter_violations_by_profile(self, violations):
        """
        Filter the violation by profile
        :param violations: 
        :return: filtered violations
        """
        profile = None

        new_violations = []

        if self.debug:
            print('violations: '+str(violations)+lineno())

        failure_count = violations['failure_count']

        if not self.profile_path:
            if self.debug:
                print('profile definition: '+str(self.profile_path)+lineno())

            try:
                profile_loader = ProfileLoader(self.custom_rule_loader.rule_definitions(),debug=self.debug)

                profile = profile_loader.load(self.custom_rule_loader.rule_definitions())
                if self.debug:
                    print('rule ids in profile: '+str(profile.rule_ids)+lineno())

            except ParserError as e:
                print('Error'+str(e)+lineno())
                sys.exit(1)
        else:
            print('has a profile path '+lineno())
            print('Need to fix for loading a profile')
            sys.exit(1)
            #profile = self.profile_path

        if profile:
            if self.debug:
                print('violations: '+str(violations)+lineno())
                print('violations: '+str(violations['violations'])+lineno())

            for v in violations['violations']:
                if self.debug:
                    print('violation: '+str(v)+lineno())

                if self.debug:
                    print('violation: '+str(v)+lineno())
                if hasattr(v,'id'):
                    if v.id:
                        if self.debug:
                            print('id: '+str(v.id)+lineno())

                        if profile.execute_rule(v.id):
                            new_violations.append(v)
        if self.debug:
            print('new_violations: '+str(new_violations)+lineno())

        results = {'failure_count':str(failure_count), 'violations':new_violations}
        return results


    def _get_schemas(self, type):
        """
        Get the schemas
        :param type: 
        :return: 
        """
        if self.debug:
            print('ValidateUtility - _get_schemas'+lineno())

        results = os.listdir(os.getcwd()+'/cloudformation_validator/test_templates/json')

        if self.debug:
            print(results)

        if type in results:
            results = os.listdir(os.getcwd() + '/cloudformation_validator/test_templates/json/'+type)
            if self.debug:
                print(results)
            return results

    def _check_structure(self, conf_structure, my_conf):
        """
        Check structure
        :param conf_structure: 
        :param my_conf: 
        :return: 
        """
        if self.debug:
            print('ValidateUtility - check structure'+lineno())
            print('type: '+str(type(conf_structure))+lineno())
            print('type: '+str(type(my_conf))+lineno())

            for k in conf_structure:
                print(k)

        if sorted(my_conf.keys()) != sorted(conf_structure.keys()):
            if self.debug:
                print('sorted keys is not equal'+lineno())
            return False

        if self.debug:
            print('keys in proper order'+lineno())

        for key in my_conf.keys():
            if self.debug:
                print('key: '+str(key))
            if type(my_conf[key]) != type(dict()):
                return False
            else:
                if sorted(my_conf[key].keys()) != sorted(conf_structure[key].keys()):
                    return False

        return True

    def _fix_JSON(self, json_message=None):
        """
        Fix JSON
        :param json_message: 
        :return: 
        """
        
        result = None
        
        try:
            result = json.loads(json_message)
        except Exception as e:
            # Find the offending character index:
            idx_to_replace = int(str(e).split(' ')[-1].replace(')', ''))
            # Remove the offending character:
            json_message = list(json_message)
            json_message[idx_to_replace] = ' '
            new_message = ''.join(json_message)
            return self._fix_JSON(json_message=new_message)
        return result

    def render_results(self, aggregate_results, output_format):
        """
        Render results
        :param aggregate_results: 
        :param output_format: 
        :return: rendered results
        """

        if self.debug:
            print('ValidateUtility - render_results'+lineno())
            print('aggregate_results: '+str(aggregate_results)+lineno())

        renderer = self.results_renderer(output_format)
        return renderer.render(aggregate_results)


    def results_renderer(self, output_format='txt'):
        """
        Sets the format for the rendered results
        :param output_format: 
        :return: 
        """

        if self.debug:
            print('ValidateUtility - results_renderer'+lineno())

        if output_format== 'txt':
            return SimpleStdoutResults()

        elif output_format=='json':
            return JsonResults(debug=self.debug,suppress_errors=self.suppress_errors)
        else:
            if not self.suppress_errors:
                print('there is a problem'+lineno())


    def count_warnings(self, violations):
        """
        Counts the number of warnings
        :param violations: 
        :return: warning count
        """
        if self.debug:
            print('count warnings'+lineno())

        if violations:
            if self.debug:
                print('length: '+str(len(violations))+lineno())

        count =0

        for violation in violations:
            if self.debug:
                print(violation)

            if violation:
                if hasattr(violation,'type'):
                    if violation.type:
                        if violation.type == 'VIOLATION::WARNING':
                            if str(violation.id) not in self.excluded_rules:
                                count+=int(len(violation.logical_resource_ids))
        return count

    def count_failures(self, violations):
        """
        Count the number of failures
        :param violations: 
        :return: returns the number of failures
        """
        if self.debug:
            print('count failures'+lineno())

        if violations:
            if self.debug:
                print('violations: '+str(violations)+lineno())
                print('length: '+str(len(violations))+lineno())


        count =0
        for violation in violations:
            if self.debug:
                print(str(violation)+lineno())

            if violation:
                if hasattr(violation,'type'):


                    if violation.type:
                        if violation.type == 'VIOLATION::FAILING_VIOLATION':

                            if str(violation.id) not in self.excluded_rules:
                                count+=int(len(violation.logical_resource_ids))

        return count


    def validate_path(self,path):

        if path:
            matchObj = re.match(r'^\./(.*)', path, re.M | re.I)
            if matchObj:

                return str(os.getcwd() + '/' + str(matchObj.group(1)))

            matchObj = re.match(r'^[a-zA-Z1-9].*', path, re.M | re.I)
            if matchObj:
                return str(os.getcwd() + '/' + str(matchObj.group(0)))

        return path


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule

def lineno():
    """Returns the current line number in our program."""
    return str(' - RDSInstanceMasterUserPasswordRule- caller: '+str(inspect.stack()[1][3])+'  - line number: '+str(inspect.currentframe().f_back.f_lineno))


class RDSInstanceMasterUserPasswordRule(BaseRule):
  
  def __init__(self, cfn_model=None, debug=None):
    """
    Initialize
    :param cfn_model: 
    """
    BaseRule.__init__(self, cfn_model, debug=debug)
      
  def rule_text(self):
    """
    Get rule text
    :return: 
    """
    if self.debug:
      print('rule_text'+lineno())
    return 'RDS instance master user password must be Ref to NoEcho Parameter. Default credentials are not recommended'


  def rule_type(self):
    """
    Get rule type
    :return: 
    """
    self.type= 'VIOLATION::FAILING_VIOLATION'
    return 'VIOLATION::FAILING_VIOLATION'

  def rule_id(self):
    """
    Get rule id
    :return: 
    """
    if self.debug:
      print('rule_id'+lineno())
    self.id ='F23'
    return 'F23'


  # one word of warning... if somebody applies parameter values via JSON.... this will compare that....
  # probably shouldn't be doing that though - if it's NoEcho there's a good reason
  # bother checking synthesized_value? that would be the indicator.....
  def audit_impl(self):
    """
    Audit
    :return: violations
    """
    if self.debug:
      print('RDSInstanceMasterUserPasswordRule - audit_impl'+lineno())
    
    violating_rdsinstances = []

    resources = self.cfn_model.resources_by_type('AWS::RDS::DBInstance')

    if len(resources)>0:
      for resource in resources:
          if self.debug:
            print('resource: '+str(resource)+lineno())
            print('vars:'+str(vars(resource))+lineno())

          if hasattr(resource,'masterUserPassword'):

            if resource.masterUserPassword:

              if not self.references_no_echo_parameter_without_default(self.cfn_model,resource.masterUserPassword):
                 violating_rdsinstances.append(str(resource.logical_resource_id))

              if self.debug:
                print('violating_rdsinstances: ' + str(violating_rdsinstances) + lineno())

    else:
      if self.debug:
        print('no violating_rdsinstances' + lineno())

    if self.debug:
      input('Press enter to continue: '+lineno())

    return violating_rdsinstances

  def to_boolean(self,string):
    """
    Why and how used
    :param string: 
    :return: 
    """
    if self.debug:
      print('to_boolean'+lineno())
    # FIXME
    sys.exit(1)
    #string.to_s.casecmp('true').zero?


  def references_no_echo_parameter_without_default(self,cfn_model, master_user_password):
    """
    ???
    :param cfn_model: 
    :param master_user_password: 
    :return: 
    """
    if self.debug:
      print('references_no_echo_parameter_without_default'+lineno())
      print('cfn_model: '+str(vars(cfn_model))+lineno())
      print('master_username: '+str(master_user_password)+lineno())

    if type(master_user_password) == type(dict()):

      if 'Ref' in master_user_password:
        if hasattr(cfn_model,'parameters'):
          if str(master_user_password['Ref']) in cfn_model.parameters:
            parameter = cfn_model.parameters[master_user_password['Ref']]

            if hasattr(parameter,'NoEcho'):
              if self.debug:
                print('has noecho property: '+lineno())
              return True

    if self.debug:
      print('Does not have noecho property '+lineno())
        
    return False


from __future__ import absolute_import, division, print_function
import copy
import inspect
import sys


def lineno():
    """Returns the current line number in our program."""
    return str(' -  Ec2InstanceParser- line number: '+str(inspect.currentframe().f_back.f_lineno))


class Ec2InstanceParser:
    """
    Ec2 instance parser
    """
    @staticmethod
    def parse(cfn_model, resource, debug=False):
        """
        Parse ec2 instances resources
        :param resource: 
        :param debug: 
        :return: 
        """

        if debug:
            print('Ec2InstancePaRSER - parse'+lineno())

        ec2_instance = copy.copy(resource)

        if hasattr(ec2_instance,'securityGroupIds'):
            if type(ec2_instance.securityGroupIds) == type(list()):

                for sg in ec2_instance.securityGroupIds:

                    if debug:
                        print('security group: '+str(sg)+lineno())
                        print('type: '+str(type(sg))+lineno())

                    if type(sg)== type(list()):
                        if debug:
                            print('security group is a list'+lineno())
                        sys.exit(1)
                    else:
                        if debug:
                            print('security group is not a list - it is a '+str(type(sg))+lineno())

                        if 'Ref' in sg:
                            ec2_instance.security_groups = cfn_model.find_security_group_by_group_id(sg['Ref'])

            else:

                print('FIXME')
                # could be a Ref to a List<AWS::EC2::SecurityGroup::Id> which we can't
                # do much with at the level of static analysis before knowing the parameter passed in
                # worth checking defaults?
                ec2_instance.security_groups = []

        return ec2_instance


from __future__ import absolute_import, division, print_function
import sys
import inspect
from builtins import (str)
from terraform_validator.RuleDefinition import RuleDefinition


def lineno():
    """Returns the current line number in our program."""
    return str(' - RuleRegistry - line number: '+str(inspect.currentframe().f_back.f_lineno))

class RuleRegistry:
    """
    Rule registry
    """
    def __init__(self, debug=False):
        """
        Initialize RulesRegistry
        :param debug: 
        """
        self.debug = debug
        self.rules = []

        if self.debug:
            print('RuleRegistry init'+lineno())


    def definition(self, id, type, message):
        """
        Add a rule definition to the registry
        :param id: 
        :param type: 
        :param message: 
        :return: 
        """
        if self.debug:
            print('definition '+lineno())

        rule_definition = RuleDefinition(id, type, message, debug=self.debug)

        if self.debug:
            print('rule_definition: '+str(rule_definition)+lineno())
            print('rule_definition id: '+str(rule_definition.id)+lineno())

        existing_def = self.by_id(rule_definition.id)

        if self.debug:
            print('existing definition: '+str(existing_def)+lineno())

        if not existing_def:
            if self.debug:
                print('adding rule '+lineno())
            self.add_rule(rule_definition)
        else:
            return existing_def



    def by_id(self, id):
        """
        Find rule definition by id
        :param id:
        :return:
        """
        if self.debug:
            print('by_id'+lineno())

        found_it = False
        if len(self.rules)>0:

            for rule in self.rules:
                if self.debug:
                  print('rule: '+str(rule))
                  print('rule id: '+str(rule.id)+lineno())


                if rule.id == str(id):
                  found_it=True
                  break

        if self.debug:
            print('found it: '+str(found_it)+lineno())

        return found_it

    def warnings(self):
        """
        Get rule warnings
        :param self:
        :return:
        """

        rules = []

        if self.debug:
            print('warnings'+lineno())

        for rule in self.rules:
            if self.debug:
                print(str(rule)+lineno())
                print('rule type: '+str(rule.type)+lineno())

            if rule.type == 'VIOLATION::WARNING':
                rules.append(rule)
        return rules

    def failings(self):
        """
        Get failing rules
        :param self:
        :return:
        """
        rules = []

        if self.debug:
            print('failings'+lineno())

        for rule in self.rules:
            if self.debug:
                print(str(rule)+lineno())
                print('rule type: '+str(vars(rule))+lineno())

            if rule.type == 'VIOLATION::FAILING_VIOLATION':
                rules.append(rule)
        return rules

    def add_rule(self, violation_def):
        """
        Add rule definition
        :param self:
        :param violation_def:
        :return:
        """
        if self.debug:
            print('add_rule'+lineno())
            print('violation_def: '+str(violation_def)+lineno())

        self.rules.append(violation_def)

        if self.debug:
            print('registry rules are now: '+str(self.rules)+lineno())



from __future__ import absolute_import, division, print_function
import inspect
import copy
import sys


def lineno():
    """Returns the current line number in our program."""
    return str(' -  LoadBalancerParser- line number: '+str(inspect.currentframe().f_back.f_lineno))


class LoadBalancerParser:
    """
    Loadbalancer parser
    """
    
    @staticmethod
    def parse(cfn_model, resource, debug=False):
        """
        Parse load balancer
        :param resource: 
        :param debug: 
        :return: 
        """
        if debug:
            print('parse'+lineno())

        load_balancer = copy.copy(resource)

        if debug:
            print('load_balancer'+str(vars(load_balancer))+lineno())
        if type(load_balancer) == type(list()):
            if debug:
                print('load_balancer is list')
            if hasattr(load_balancer,'securityGroups'):
                for sgs in load_balancer.securitygroups:
                    if debug:
                        print(sgs)
                    # FIXME
                    #cfn_model.find_security_group_by_group_id(security_group_reference
                    sys.exit(1)
        else:
            if debug:
                print('load_balancer is not a list'+lineno())
            load_balancer.security_groups= []
            
        return load_balancer


from __future__ import absolute_import, division, print_function
import os
import sys
import inspect

def lineno():
    """Returns the current line number in our program."""
    return str(' - TemplateDiscovery - line number: '+str(inspect.currentframe().f_back.f_lineno))


class TemplateDiscovery:
    """
    Template discover
    """

    def __init__(
            self, 
            input_json_path=str(os.path.dirname(__file__)+'/test_templates/json'),
            template_pattern='json', 
            debug=False
    ):
        """
        Initialize TemplateDiscovery
        :param input_json_path: 
        :param template_pattern: 
        :param debug: 
        """
        self.input_json_path = input_json_path
        self.template_pattern = template_pattern
        self.templates = []
        self.debug = debug

        if self.debug:
            print('TemplateDiscovery - __init__'+lineno())

    def discover_templates(self):
        """
        Discover templates
        :return: 
        """
        if self.debug:
            print("\n\n#######################################")
            print('discover templates'+lineno())
            print("###########################################\n\n")

            print('input_json_path: '+str(self.input_json_path)+lineno())
            print('template_pattern: '+str(self.template_pattern)+lineno())

            print(str(os.path.dirname(__file__)))
            print('cwd: '+str(os.chdir(os.path.dirname(__file__)))+lineno())

        template_filenames = []

        for root, directories, filenames in os.walk(self.input_json_path):
            for filename in filenames:
                if filename.endswith('.json') or filename.endswith('.tf'):

                    if self.debug:
                        print(os.path.join(root, filename))

                    #if 'AWSTemplateFormatVersion' in open(os.path.join(root,filename)).read():
                    template_filenames.append(os.path.join(root, filename))

        if self.debug:
            print('file names: '+str(template_filenames)+lineno())


        return template_filenames

    def render_path(self):
        """
        Render the path
        :return: 
        """
        if self.debug:
            print('render_path'+lineno())
        # FIXME
        sys.exit(1)
        return path
        # return path.path if path.is_a? File
        # path

    def find_templates_in_directory(self):
        """
        Find templates in directory
        :return: 
        """
        if self.debug:
            print('find_template_in_directory'+lineno())
        # FIXME
        sys.exit(1)



import os
import sys
import unittest
from collections import OrderedDict
from terraform_model.model.EC2NetworkInterface import EC2NetworkInterface as class_to_test

def pretty(value, htchar='\t', lfchar='\n', indent=0):
    nlch = lfchar + htchar * (indent + 1)
    if type(value) == type(dict()) or type(value) == type(OrderedDict()):

        items = [
            nlch + repr(key) + ': ' + pretty(value[key], htchar, lfchar, indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) == type(list()):
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '[%s]' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) is tuple:
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + lfchar + htchar * indent)
    else:
        return repr(value)

class TestEc2Interface(unittest.TestCase):
    """
    Test ipaddr
    """

    def test_new_interface(self):


      cfn_model = {

          "Resources": {
              "myENI": {
                  "Type": "AWS::EC2::NetworkInterface",
                  "Properties": {
                      "Tags": [{"Key": "foo", "Value": "bar"}],
                      "Description": "A nice description.",
                      "SourceDestCheck": "false",
                      "GroupSet": ["sg-75zzz219"],
                      "SubnetId": "subnet-3z648z53",
                      "PrivateIpAddress": "10.0.0.16"
                  }
              }
          }
        }



      expected_result = 'EC2NetworkInterface'

      real_result = class_to_test(cfn_model=cfn_model)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))
      print('type: '+str(type(real_result)))
      print('name: '+str(real_result.__class__.__name__))
      self.maxDiff = None
      self.assertEqual(expected_result, real_result.__class__.__name__)




from __future__ import absolute_import, division, print_function
import inspect
import sys
import time
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule
from terraform_validator.IpAddr import IpAddr


def lineno():
    """Returns the current line number in our program."""
    return str(' - SecurityGroupIngressOpenToWorldRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))


class SecurityGroupIngressOpenToWorldRule(BaseRule):
  

    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Get rule text
        :return:
        """
        if self.debug:
            print('rule_text'+lineno())
        return 'Security Groups found with cidr open to world on ingress.  This should never be true on instance.  Permissible on ELB'


    def rule_type(self):
        """
        Get rule type
        :return:
        """
        self.type= 'VIOLATION::WARNING'
        return 'VIOLATION::WARNING'


    def rule_id(self):
        """
        Get rule id
        :return:
        """
        if self.debug:
            print('rule_id'+lineno())
        self.id ='W2'
        return 'W2'



    def audit_impl(self):
        """
        Audit
        :return: violations
        """
        if self.debug:
            print('SecurityGroupIngressOpenToWorldRule - audit_impl'+lineno())
        violating_ingresses = []

        for groups in self.cfn_model.security_groups():
            if self.debug:
                print('group: '+str(groups)+lineno())
                print('vars: '+str(vars(groups))+lineno())

            for ingress in groups.ingresses:
                if self.debug:
                    print('ingress: '+str(ingress)+lineno())
                    print('type: '+str(type(ingress)))

                if type(ingress)== type(dict()):
                    if IpAddr.ip4_open(ingress,debug=self.debug) or IpAddr.ip6_open(ingress,debug=self.debug):
                        if self.debug:
                            print('ip4/6 address is open'+lineno())
                        violating_ingresses.append(str(groups.logical_resource_id))
                else:
                    if IpAddr.ip4_open(ingress,debug=self.debug) or IpAddr.ip6_open(ingress,debug=self.debug):
                        if self.debug:
                            print('ip4/6 address is open'+lineno())
                        violating_ingresses.append(str(ingress.logical_resource_id))

        routes= self.cfn_model.standalone_ingress()

        if self.debug:
            print('routes: '+str(routes)+lineno())
        for standalone_ingress in routes:
            if self.debug:
                print('standalone_ingress: '+str(standalone_ingress)+lineno())
                print('vars: '+str(vars(standalone_ingress))+lineno())

            if IpAddr.ip4_open(standalone_ingress,debug=self.debug) or IpAddr.ip6_open(standalone_ingress,debug=self.debug):
                if self.debug:
                    print('ip4/6 address is open' + lineno())
                violating_ingresses.append(str(standalone_ingress.logical_resource_id))

        if self.debug:
            input('Press enter to continue: ' + lineno())

        return violating_ingresses


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule
from collections import Iterable
from six import StringIO, string_types
from builtins import (str)


def lineno():
    """Returns the current line number in our program."""
    return str(' -  S3HasTagsRule - caller: ' + str(inspect.stack()[1][3]) + ' - line number: ' + str(
        inspect.currentframe().f_back.f_lineno))

class S3CustomTagsRule(BaseRule):
    """
    S3 custom tags rule
    """
    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize Ec2HasTagsRule
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Returns rule text
        :return:
        """
        if self.debug:
            print('rule_text' + lineno())
        return 'S3 bucket does not have the required tags of Name, ResourceOwner, DeployedBy, Project'

    def rule_type(self):
        """
        Returns rule type
        :return:
        """
        self.type = 'VIOLATION::FAILING_VIOLATION'
        return 'VIOLATION::FAILING_VIOLATION'

    def rule_id(self):
        """
        Returns rule id
        :return:
        """
        if self.debug:
            print('rule_id' + lineno())
        self.id = 'F87'
        return 'F87'

    def tags_to_dict(self, aws_tags):
        """ Convert a list of AWS tags into a python dict """
        return {str(tag['Key']): str(tag['Value']) for tag in self.ensure_list(aws_tags)}

    def ensure_list(self, value):
        """
        Coerces a variable into a list; strings will be converted to a singleton list,
        and `None` or an empty string will be converted to an empty list.
        Args:
            value: a list, or string to be converted into a list.

        Returns:
            :py:class:`list`
        """
        ret_value = value
        if not value:
            ret_value = []
        elif not isinstance(value, Iterable) or isinstance(value, string_types):
            ret_value = [value]
        return ret_value

    def audit_impl(self):
        """
        Audit
        :return: violations
        """
        if self.debug:
            print('S3HasTagsRule - audit_impl' + lineno())

        violating_volumes = []

        resources = self.cfn_model.resources_by_type('AWS::S3::Bucket')

        if len(resources) > 0:

            for resource in resources:
                if self.debug:
                    print('resource: ' + str(resource) + lineno())
                    print('vars: ' + str(vars(resource)))

                if hasattr(resource, 'tags'):
                    if self.debug:
                        print('has tags attribute' + lineno())

                    tags_dict = self.tags_to_dict(resource.cfn_model['Properties']['Tags'])

                    required_tags = ('Name', 'ResourceOwner', 'DeployedBy', 'Project')
                    if not set(required_tags).issubset(tags_dict):
                        violating_volumes.append(str(resource.logical_resource_id))
                else:
                    if self.debug:
                        print('does not tags property')
                    violating_volumes.append(str(resource.logical_resource_id))

        else:
            if self.debug:
                print('no violating_volumes' + lineno())

        return violating_volumes



from __future__ import absolute_import, division, print_function
from builtins import (str)
import sys

class LambdaPrincipal:
    """
    Lambda principal model
    """
    
    @staticmethod
    def wildcard(context):
        """
        If contains wildcard
        :return: 
        """
        if type(context) == type(str()):
            if '*' in context:
                return True

        if sys.version_info[0] < 3:
            if type(context) == type(unicode()):
                if '*' in context:
                    return True

        return False



from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule

def lineno():
    """Returns the current line number in our program."""
    return str(' - SqsQueuePolicyWildcardPrincipalRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))


class SqsQueuePolicyWildcardPrincipalRule(BaseRule):
  
    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Get rule text
        :return:
        """
        if self.debug:
            print('rule_text'+lineno())
        return 'SQS Queue policy should not allow * principal'

    def rule_type(self):
        """
        Get rule type
        :return:
        """
        self.type= 'VIOLATION::FAILING_VIOLATION'
        return 'VIOLATION::FAILING_VIOLATION'

    def rule_id(self):
        """
        Get rule id
        :return:
        """
        if self.debug:
            print('rule_id'+lineno())
        self.id ='F21'
        return 'F21'

    def audit_impl(self):
        """
        Audit
        :return: violations
        """
        if self.debug:
            print('SqsQueuePolicyWildcardPrincipalRule - audit_implt'+lineno())
        violating_rules = []

        resources = self.cfn_model.resources_by_type('AWS::SQS::QueuePolicy')

        if len(resources)>0:
            for resource in resources:
                if self.debug:
                    print('resource: '+str(resource)+lineno())

                if hasattr(resource, 'policy_document'):
                    if resource.policy_document:
                        if self.debug:
                            print('policy document: ' + str(resource.policy_document))
                        if resource.policy_document.wildcard_allowed_principals():

                            violating_rules.append(str(resource.logical_resource_id))
        else:
            if self.debug:
                print('no violating_rules' + lineno())

        if self.debug:
            input('Press enter to continue: ' + lineno())

        return violating_rules


from __future__ import absolute_import, division, print_function
import inspect
import sys
from terraform_model.parser.PolicyDocumentParser import PolicyDocumentParser
from terraform_model.model.Policy import Policy


def lineno():
    """Returns the current line number in our program."""
    return str(' -  IamUserParser- line number: '+str(inspect.currentframe().f_back.f_lineno))

class IamUserParser:
    """
    IAM User Parser
    """

    @staticmethod
    def parse(cfn_model, resource, debug=False):
        """
        Parse iam user
        :param resource: 
        :param debug: 
        :return: 
        """

        if debug:
            print('parse'+lineno())

        iam_user = resource

        for policy in iam_user.policies:
            if debug:
                print('policy: '+str(policy)+lineno())

            new_policy = Policy(debug=debug)
            new_policy.policy_name = policy['PolicyName']
            policy_document_parser =PolicyDocumentParser()
            new_policy.policy_document=policy_document_parser.parse(policy['PolicyDocument'])

            iam_user.policy_objects.append(new_policy)

        for group_name in iam_user.groups:
            if debug:
                print('group_name: '+str(group_name)+lineno())
            iam_user.group_names.append(group_name)

        user_to_group_additions = cfn_model.resources_by_type('AWS::IAM::UserToGroupAddition')

        if debug:
            print('user_to_group_additions: '+str(user_to_group_additions)+lineno())

        for user_to_group_addition in user_to_group_additions:
            if IamUserParser.user_to_group_addition_has_username(user_to_group_addition.users,iam_user,debug=debug):
                iam_user.group_names = user_to_group_addition.groupName
                #    # we need to figure out the story on resolving Refs i think for this to be real

        return iam_user



    def user_to_group_addition_has_username(addition_user_names, user_to_find, debug=False):
        """
        ???
        :param user_to_find: 
        :param debug: 
        :return: 
        """
        if debug:
            print('user_to_group_addition_has_username'+lineno())
            print('user names: '+str(addition_user_names)+lineno())
            print('user to find: '+str(user_to_find)+lineno())

        for addition_user_name in addition_user_names:

            if debug:
                print('vars: '+str(vars(user_to_find))+lineno())

            if hasattr(user_to_find,'userName'):
                if addition_user_name == user_to_find.userName:
                    return True

            if type(addition_user_name) == type(dict()):
                if 'Ref' in addition_user_name:
                    if addition_user_name['Ref'] == user_to_find.logical_resource_id:
                        return True

        return False


from __future__ import absolute_import, division, print_function
from terraform_model.model.ModelElement import ModelElement

class SNSTopicPolicy(ModelElement):
    """
    SNS Topic Policy model
    """

    _cfn_model = None

    def __init__(self, cfn_model):
        """
        Initialize
        :param cfn_model:
        """

        ModelElement.__init__(self, cfn_model)

        self._topics = []
        self.policy_document = None
        self._resource_type = 'AWS::SNS::TopicPolicy'





from __future__ import absolute_import, division, print_function
import sys
import os
import copy
import importlib
import inspect
import boto3
from botocore.exceptions import ClientError
from terraform_validator.RuleRegistry import RuleRegistry

def lineno():
    """Returns the current line number in our program."""
    return str(' - CustomRuleLoader - line number: '+str(inspect.currentframe().f_back.f_lineno))

def dump(obj):
  for attr in dir(obj):
    print("obj.%s = %r" % (attr, getattr(obj, attr)))

class CustomRuleLoader:
    """
    Custom rule loader
    """
    def __init__(self, debug=False,
                 rule_directory=None,
                 extra_rule_directory=None,
                 allow_suppression=True,
                 print_suppression=False,
                 isolate_custom_rule_exceptions=False,
                 additional_rules_directory=None,
                 excluded_rules=[],
                 s3_bucket_name=None,
                 s3_profile = None,
                 temp_dir_path=None):
        """
        Initialize CustomRuleLoader
        :param debug: 
        :param rule_directory: 
        :param allow_suppression: 
        :param print_suppression: 
        :param isolate_custom_rule_exceptions: 
        """
        self.debug = debug
        self.rule_directory = rule_directory
        self.extra_rule_directory = extra_rule_directory
        self.additional_rules_directory = additional_rules_directory
        self.allow_suppression= allow_suppression
        self.print_suppression = print_suppression
        self.isolate_custom_rule_exceptions = isolate_custom_rule_exceptions
        self.validate_extra_rule_directory = self.rule_directory
        self.excluded_rules = excluded_rules
        self.s3_bucket_name = s3_bucket_name
        self.temp_dir_path= temp_dir_path
        self.s3_profile = s3_profile

        if self.debug:
            print('CustomRuleLoader init'+lineno())

    def rule_definitions(self):
        """
        Loads all the rule definitions from the rules directory in to the rules registry object and
        returns the rules registry object
        :return: rules registry
        """
        if self.debug:
            print('CustomRuleLoader - rule definitions'+str(lineno()))

        rule_registry= RuleRegistry(self.debug)


        classes = self.discover_rule_classes(self.rule_directory)

        if self.debug:
            print('number of rule types: '+str(len(classes))+lineno())

        # Iterate through the rules directory and get the class for each of the rules
        for rule_class in classes:
            if self.debug:
                print(str(rule_class)+lineno())

            for file in classes[rule_class]:

                # Skip the base rule from which all rules are derived
                if str(file)== "BaseRule":
                    continue

                id, type,message = self.rule_registry_from_rule_class(file, classes[rule_class][file])

                if self.debug:
                    print('id: '+str(id)+lineno())
                    print('type: '+str(type)+lineno())
                    print('message: '+str(message)+lineno())

                if str(id) not in self.excluded_rules:
                    rule_registry.definition(id, type, message)

        if self.debug:
            print('rules registry rules are now: '+str(rule_registry.rules)+lineno())

        return rule_registry

    def execute_custom_rules(self, cfn_model):
        """
        Execute any custom rules, which are in addition to the regular rules
        :param cfn_model:
        :return:
        """
        if self.debug:
            print('CustomRuleLoader - execute_custom_rules'+str(lineno()))
            print('cfn_model: '+str(cfn_model)+str(lineno()))

        violations = []

        # Validate the cfn model again the metadata nagger
        cfn_model  = self.validate_cfn_nag_metadata(cfn_model)

        if self.debug:
            print('##### Done checking metadata #########'+str(lineno()))
            print('### filtering rule classes'+str(lineno()))

        violations = self.filter_rule_classes(cfn_model, violations)

        if self.debug:
            print('violations from extra rules: '+str(violations)+lineno())

        return violations

    def rule_registry_from_rule_class(self, rule_class, directory):
        """
        Gets the rule id, type and text for a specific rule
        :param rule_class: The rule name
        :return: rule_id, rule_type, rule_text
        """
        if self.debug:
            print('CustomRuleLoader - rule_registry_from_rule_class'+str(lineno()))
            print('rule_class: '+str(rule_class)+lineno())
            print('rule directory: '+str(directory))

        if 'site-packages' in str(directory) and 'additional' in str(directory):
            my_module = importlib.import_module("terraform_validator.additional_custom_rules." + str(rule_class))
        elif 'site-packages' in str(directory) and 'custom_rules' in str(directory):
            my_module = importlib.import_module("terraform_validator.custom_rules." + str(rule_class))
        elif 'terraform_validator' in str(directory) and 'additional' in str(directory):
            my_module = importlib.import_module("terraform_validator.additional_custom_rules." + str(rule_class))
        elif 'terraform_validator' in str(directory) and 'custom_rules' in str(directory):
            my_module = importlib.import_module("terraform_validator.custom_rules." + str(rule_class))
        else:
            my_module = importlib.import_module(str(rule_class))

        MyClass = getattr(my_module, str(rule_class))

        if self.debug:
            print('Created new class:'+str(rule_class) + lineno())
            print('type: '+str(type(MyClass))+lineno())
            print('vars: '+str(vars(MyClass))+lineno())

        rule = MyClass()

        if self.debug:
            print('rule: '+str(rule)+lineno())
            print('type: '+str(type(rule))+lineno())
            print('vars: '+str(vars(rule))+lineno())
            print(str(rule.rule_id()) + lineno())

        return rule.rule_id(), rule.rule_type(), rule.rule_text()

    def filter_rule_classes(self, cfn_model, violations):
        """
        Filter rules
        :param cfn_model: 
        :param violations: 
        :return: violations
        """
        if self.debug:
            print('CustomRuleLoader - filter_rule_classes'+str(lineno()))

        # Get an array of all the rule classes
        rule_classes = self.discover_rule_classes(self.rule_directory)


        # Iterate over each of the rule classes
        for rule_class in rule_classes:

            for file in rule_classes[rule_class]:

                if self.debug:
                    print('file: '+str(file)+lineno())

                # Ignore the base rule
                if file == 'BaseRule':
                    continue

                if rule_class == 'rules_directory':

                    if self.debug:
                        print("\n\n##########################")
                        print('Rule is not BaseRule'+lineno())
                        print('CustomRuleLoader - filter_rule_classes - rule_class '+str(file)+str(lineno()))
                        print("##########################\n\n")

                    # Import the rule class
                    if 'additional' in rule_classes[rule_class][file]:
                        my_module = importlib.import_module("terraform_validator.additional_custom_rules." + str(file))
                    else:
                        my_module = importlib.import_module("terraform_validator.custom_rules."+str(file))

                elif rule_class == 'extra_rules_directory':

                    if self.debug:
                        print("\n\n##########################")
                        print('Rule is not BaseRule' + lineno())
                        print('CustomRuleLoader - filter_rule_classes - rule_class ' + str(file) + str(lineno()))
                        print("##########################\n\n")

                    # Import the rule class
                    my_module = importlib.import_module(str(file))

                MyClass = getattr(my_module, str(file))

                if self.debug:
                    print('debug is currently: '+str(self.debug)+lineno())


                if self.debug:
                    print("\n\n#############################################")
                    print('Created new '+str(file)+' class'+lineno())
                    print('rule class: '+str(file)+lineno())
                    print('my class: '+str(type(MyClass))+lineno())
                    print("################################################\n")

                # Sets the cfn_model in the rule class
                instance = MyClass(cfn_model, debug=self.debug)

                if self.debug:
                    print(str(instance.rule_id())+lineno())
                    print(dir(instance))

                    print("\n\n############################")
                    print('filtering the cfn_model to remove suppressed resources'+lineno())
                    print("#################################\n\n")

                # This does nothing if terraform_validator allow_suppression flag is False
                filtered_cfn_model = self.cfn_model_with_suppressed_resources_removed(
                    cfn_model=cfn_model,
                    rule_id=instance.rule_id(),
                    allow_suppression=self.allow_suppression
                )

                # Set the new cfn_model to the filtered cfn_model
                instance.cfn_model = copy.copy(filtered_cfn_model)

                if self.debug:
                    print("\n\n#####################################################")
                    print('Auditing rule id: '+str(instance.rule_id())+lineno())
                    print('rule text: '+str(instance.rule_text())+lineno())
                    print('calling instance.audit '+lineno())
                    print("########################################################\n")

                # The rule is a subclass of the BaseRule.  This calls the BaseRule audit function
                audit_result = instance.audit()

                if self.debug:
                    print('audit results: '+str(audit_result)+lineno())

                # Prints audit results to stdout
                if audit_result:
                    if self.debug:
                        print(str(dir(audit_result)) + lineno())
                        print(str(vars(audit_result))+lineno())
                        print(str(audit_result.message)+lineno())

                    # Adds audit results to violation
                    violations.append(audit_result)

        return violations

    def rules_to_suppress(self, resource):
        """
        Rules to suppress
        :param resource: 
        :return: 
        """
        if self.debug:
            print('CustomRuleLoader - rules_to_suppress'+str(lineno()))
            print('resource: '+str(resource)+lineno())
            print('vars: '+str(vars(resource))+lineno())

        if hasattr(resource, 'metadata'):
            if resource.metadata:
                if 'cfn_validator' in resource.metadata:
                    if 'rules_to_suppress' in resource.metadata['cfn_validator']:
                        return resource.metadata['cfn_validator']['rules_to_suppress']
        else:
            if self.debug:
                print('no rules to suppress'+str(lineno()))

            return None


    def validate_cfn_nag_metadata(self, cfn_model):
        """
        Validate the metadata
        :param cfn_model: 
        :return: 
        """
        if self.debug:
            print('CustomRuleLoader - validate_cfn_nag_metadata'+str(lineno()))
            print('cfn_model: '+str(cfn_model)+lineno())

        mangled_metadatas = []

        for resource in cfn_model.resources:
            if self.debug:
                print('resource: '+str(resource)+lineno())
            resource_id = resource
            resource_rules_to_suppress = self.rules_to_suppress(cfn_model.resources[resource])

            if resource_rules_to_suppress:
                for rules in resource_rules_to_suppress:
                    if self.debug:
                        print('rule:'+str(rules)+lineno())

        return cfn_model


    def suppress_resource(self, rules_to_suppress, rule_id, logical_resource_id):
        """
        Supress certain resources
        :param rules_to_suppress: 
        :param rule_id: 
        :param logical_resource_id: 
        :return: 
        """
        if self.debug:
            print('CustomRuleLoader - suppress_resource'+str(lineno()))

        sys.exit(1)
        #FIXME
        found_suppression_ruile = self.rules_to_suppress()
        # found_suppression_rule = rules_to_suppress.find do |rule_to_suppress|
        #   next if rule_to_suppress['id'].nil?
        #   rule_to_suppress['id'] == rule_id
        # end
        # if found_suppression_rule && @print_suppression
        #   STDERR.puts "Suppressing #{rule_id} on #{logical_resource_id} for reason: #{found_suppression_rule['reason']}"
        # end
        # !found_suppression_rule.nil?


    def cfn_model_with_suppressed_resources_removed(self, cfn_model, rule_id, allow_suppression):
        """
        CFN model with suppressed resources removed
        :param cfn_model: 
        :param rule_id: 
        :param allow_suppression: 
        :return: 
        """
        if self.debug:
            print('CustomRuleLoader - cfn_model_with_suppressed_resources_removed'+str(lineno()))


        # Whether terraform_validator allow suppression is False
        if not allow_suppression:
            if self.debug:
                print('not allow suppression '+lineno())

            return cfn_model

        # If we are allowing suppression of certain rules
        else:
            if self.debug:
                print("\n####################################")
                print('Iterating through resources in cf model '+lineno())
                print('allow suppression '+lineno())
                print('suppression is: '+str(allow_suppression))
                print("#######################################\n")
            cfn_model = copy.copy(cfn_model)

            for resource in cfn_model.resources:
                if self.debug:
                    print("resource name: "+str(resource)+lineno())

                rules_to_suppress = self.rules_to_suppress(cfn_model.resources[resource])

                # Return the model if there are no rules to suppress
                if rules_to_suppress == None:
                    if self.debug:
                        print('no rules to suppress '+lineno())
                    return cfn_model
                else:
                    if self.debug:
                        print('there are rules to suppress'+lineno())
                    self.suppress_resource(rules_to_suppress,rule_id,resource)

            return cfn_model

    def validate_extra_rule_directory(self, rule_directory):
        """
        Validate the extra rules directory
        :param rule_directory: 
        :return: 
        """
        if self.debug:
            print('CustomRuleLoader - validate_extra_rule_directory'+str(lineno()))
            print('rule_directory: '+str(rule_directory)+lineno())

        if rule_directory == None:
            return True
        else:
            return os.path.isdir(rule_directory)

    def download_dir(self, client, resource, dist, local='/tmp', bucket='your_bucket'):
        paginator = client.get_paginator('list_objects')
        for result in paginator.paginate(Bucket=bucket, Delimiter='/', Prefix=dist):
            if result.get('CommonPrefixes') is not None:
                for subdir in result.get('CommonPrefixes'):
                    download_dir(client, resource, subdir.get('Prefix'), local, bucket)
            if result.get('Contents') is not None:
                for file in result.get('Contents'):
                    if not os.path.exists(os.path.dirname(local + os.sep + file.get('Key'))):
                        os.makedirs(os.path.dirname(local + os.sep + file.get('Key')))
                    resource.meta.client.download_file(bucket, file.get('Key'), local + os.sep + file.get('Key'))



    def discover_rule_filenames(self):
        """
        Discover the rule filenames
        :return: A dictionary with the rules
        """
        if self.debug:
            print('CustomRuleLoader - discover_rule_filename'+str(lineno()))

        rule_filenames = {
            'rules_directory':{},
            'extra_rules_directory':{}
        }

        temp_rule_filenames = os.listdir(self.rule_directory)
        if self.debug:
            print('rules directory: '+str(self.rule_directory)+lineno())

        for temp_file in temp_rule_filenames:

            if self.debug:
                print('temp file: '+str(temp_file)+lineno())

            if not temp_file.startswith('__'):
                if temp_file.endswith('.py'):
                    rule_filenames['rules_directory'][str(temp_file).replace('.py','')] = str(self.rule_directory)

        if self.additional_rules_directory:
            temp_rule_filenames = os.listdir(self.additional_rules_directory)
            if self.debug:
                print('additional rules directory: ' + str(self.additional_rules_directory) + lineno())

            for temp_file in temp_rule_filenames:

                if self.debug:
                    print('temp file: ' + str(temp_file) + lineno())

                if not temp_file.startswith('__'):
                    if temp_file.endswith('.py'):
                        rule_filenames['rules_directory'][str(temp_file).replace('.py', '')] = str(self.additional_rules_directory)

        if self.extra_rule_directory:
            if self.debug:
                print('there is an extra rule directory '+lineno())

            temp_rule_filenames = os.listdir(self.extra_rule_directory)
            if self.debug:
                print('rules directory: ' + str(self.extra_rule_directory) + lineno())

            for temp_file in temp_rule_filenames:
                if not temp_file.startswith('__'):

                    if temp_file.endswith('.py'):

                        new_filename = temp_file.replace('.py','')

                        if self.debug:
                            print('new filename: '+str(new_filename)+lineno())

                        rule_filenames['extra_rules_directory'][str(temp_file).replace('.py', '')] = str(self.extra_rule_directory)

        if self.s3_bucket_name:

            try:

                if self.s3_profile:
                    if self.debug:
                        print('trying to create boto3 session to: '+str(self.s3_profile))
                    session = boto3.Session(profile_name=str(self.s3_profile))

                    # Any clients created from this session will use credentials
                    # from the [dev] section of ~/.aws/credentials.
                    s3 = session.resource('s3')

                else:
                    s3 = boto3.resource('s3')

                # select bucket
                my_bucket = s3.Bucket(self.s3_bucket_name)

                if self.debug:
                    print('bucket: '+str(self.s3_bucket_name))
                    print('tmp directory: '+str(self.temp_dir_path))

                # download file into current directory
                for object in my_bucket.objects.all():
                    my_bucket.download_file(object.key, os.path.join(self.temp_dir_path, object.key))

            except ClientError as e:
                print('Error getting custom rules from S3 bucket.  Check your profile and bucket name')
                print(str(e))
                sys.exit(1)

            # Get all the files downloaded from S3 to the temp directory
            temp_rule_filenames = os.listdir(self.temp_dir_path)

            if self.debug:
                print('rules directory: ' + str(self.temp_dir_path) + lineno())

            for temp_file in temp_rule_filenames:
                if not temp_file.startswith('__'):

                    if temp_file.endswith('.py'):

                        new_filename = temp_file.replace('.py', '')

                        if self.debug:
                            print('new filename: ' + str(new_filename) + lineno())

                        rule_filenames['extra_rules_directory'][str(temp_file).replace('.py', '')] = str(
                            self.temp_dir_path)


        if self.debug:
            print('rule filenames: '+str(rule_filenames)+lineno())

        return rule_filenames

    def discover_rule_classes(self, rule_directory):
        """
        Discover the rule classes
        :param rule_directory: 
        :return: 
        """
        if self.debug:
            print('CustomRuleLoader - discover_rule_classes'+str(lineno()))
            print('rule_directory: '+str(rule_directory))

        rule_classes = []

        rule_filenames = self.discover_rule_filenames()

        for directories in rule_filenames:

            if self.debug:
                print('directories: '+str(directories)+lineno())

            if directories == 'rules_directory':
                for file in rule_filenames['rules_directory']:

                    if self.debug:
                        print('file: '+str(file)+lineno())
                        print('directory: '+str(rule_filenames[directories][file]))

                    if 'additional' in str(rule_filenames[directories][file]):
                        exec("from terraform_validator.additional_custom_rules import "+str(file))
                    else:
                        exec("from terraform_validator.custom_rules import "+str(file))

            if directories == 'extra_rules_directory':
                import sys

                for file in rule_filenames['extra_rules_directory']:

                    sys.path.insert(0, rule_filenames['extra_rules_directory'][file])
                    exec("import "+str(file))

        if self.debug:
            print('rule filenames: '+str(rule_filenames)+lineno())

        return rule_filenames
